from . import *

def genG(*args):
    return gen_graph(*args)

# generate a connected Watts-Strogatz small-world graph
# (n,k,p) = (number of nodes, each node connected to k-nearest neighbors, probability of rewiring)
# k has to be even, tries is number of attempts to make connected graph
def gen_graph(tg, seed=None):
    if tg.graphtype=="wattsstrogatz":
        g=nx.connected_watts_strogatz_graph(tg.numnodes,tg.numlinks,tg.prob_rewire,1000,seed) # networkx graph
    elif tg.graphtype=="random":                               
        g=nx.erdos_renyi_graph(tg.numnodes, tg.prob_rewire,seed)
    elif tg.graphtype=="steyvers":
        g=genSteyvers(tg.numnodes, tg.numlinks, seed=seed, tail=tg.steyvers_tail)
    
    a=np.array(nx.to_numpy_matrix(g)).astype(int)
    return g, a

def genSteyvers(n,m, tail=True, seed=None):               # tail allows m-1 "null" nodes in neighborhood of every node
    nplocal=np.random.RandomState(seed)
    a=np.zeros((n,n))                                  # initialize matrix
    for i in range(m):                                 # complete m x m graph
        for j in range(m):
            if i!= j:
                a[i,j]=1
    for i in range(m,n):                               # for the rest of nodes, preferentially attach
        nodeprob=sum(a)/sum(sum(a))                    # choose node to differentiate with this probability distribution
        diffnode=nplocal.choice(n,p=nodeprob)        # node to differentiate
        h=list(np.where(a[diffnode])[0]) + [diffnode]  # neighborhood of diffnode
        if tail==True:
            h=h + [-1]*(m-1)
        #hprob=sum(a[:,h])/sum(sum(a[:,h]))                 # attach proportional to node degree?
        #tolink=nplocal.choice(h,m,replace=False,p=hprob)
        tolink=nplocal.choice(h,m,replace=False)          # or attach randomly
        for j in tolink:
            if j != -1:
                a[i,j]=1
                a[j,i]=1
    return nx.to_networkx_graph(a)





# create dicts for passing instead of passing many variables
# fill in missing variables with defaults

from . import *

class Data():
    
    def __init__(self, data):
        
        self.rawdata = data
    
        def generateGroupSpace():
            items = data['items']
            groupitems={}
            idx=0
            for subitems in items:
                for item in list(items[subitems].values()):
                    if item not in list(groupitems.values()):
                        groupitems[idx] = item
                        idx += 1
            return groupitems
        
        self.groupitems = generateGroupSpace()
        self.groupnumnodes = len(self.groupitems)
        self.subs = sorted(data['Xs'].keys())
        self.listnums = [(sub, listnum) for sub in self.subs for listnum in sorted(self.rawdata['Xs'][sub].keys())]
        
        if data['structure']:
            self.structure = "hierarchical"
            self.hierarchical()
        else:
            self.structure = "nonhierarchical"
            self.nonhierarchical()
        return

    def hierarchical(self):

        Xs = copy.deepcopy(self.rawdata['Xs'])
        self.Xs = [[Xs[i][j] for j in sorted(Xs[i].keys())] for i in self.subs]

        items = copy.deepcopy(self.rawdata['items'])
        self.items = [items[i] for i in sorted(items.keys())]

        self.labeledXs = [numToItemLabel(self.Xs[i],self.items[i]) for i in range(len(self.Xs))]

        try:
            irts = copy.deepcopy(self.rawdata['irts'])
            self.irts = [[irts[i][j] for j in sorted(irts[i].keys())] for i in self.subs]
        except:
            self.irts = []

        self.numnodes = [len(i) for i in self.items]
        self.structure = "hierarchical"

        return self

    def nonhierarchical(self):
        # map everyone to group space
        reverseGroup = reverseDict(self.groupitems)
        Xs = copy.deepcopy(self.rawdata['Xs'])
        items = copy.deepcopy(self.rawdata['items'])
        irts = copy.deepcopy(self.rawdata['irts'])

        for sub in Xs:
            for listnum in Xs[sub]:
                Xs[sub][listnum] = [reverseGroup[items[sub][i]] for i in Xs[sub][listnum]]
                
        try:
            self.Xs = flatten_list([[Xs[i][j] for j in sorted(Xs[i].keys())] for i in self.subs])
            self.irts = flatten_list([[irts[i][j] for j in sorted(irts[i].keys())] for i in self.subs])
        except:
            self.irts = []

        self.numnodes = self.groupnumnodes
        self.items = self.groupitems
        self.structure = "nonhierarchical"
        self.labeledXs = numToItemLabel(self.Xs, self.items)
        
        return self

    def subs(self):
        return 
    
        
def DataModel(data):
    tdkeys=list(data.keys())

    # full factorial of any list params
    for i in tdkeys:
        if isinstance(data[i],list):
            return flatten_list([Data(dict(data, **{i: j})) for j in data[i]])

    if 'trim' not in tdkeys:
        data['trim'] = 1.0           # each list covers full graph by default
    if 'jump' not in tdkeys:
        data['jump'] = 0.0           # no jumping in data by default
    if 'jumptype' not in tdkeys:
        data['jumptype'] = "uniform"   # or stationary
    if 'start_node' not in tdkeys:
        data['start_node'] = "stationary"      # or stationary
    #if 'numx' not in tdkeys:
    #    raise ValueError("Must specify 'numx' in data!")
    if 'priming' not in tdkeys:
        data['priming'] = 0.0
    if 'jumponcensored' not in tdkeys:
        data['jumponcensored'] = None
    if 'censor_fault' not in tdkeys:    # for generative process only, not inferring graphs
        data['censor_fault'] = 0.0
    if 'emission_fault' not in tdkeys:  # for generative process only, not inferring graphs
        data['emission_fault'] = 0.0
    
    return dotdict(data)

def Irts(irts):
    irtkeys=list(irts.keys())

    # full factorial of any list params
    for i in irtkeys:
        if (isinstance(irts[i],list)) and (i != 'data'):       # 'data' is an exception to the rule
            return flatten_list([Irts(dict(irts, **{i: j})) for j in irts[i]])

    if 'data' not in irtkeys:
        irts['data']=[]

    if 'irttype' not in irtkeys:
        if len(irts['data']) > 0:        # error unless empty dict (no IRTs)
            raise ValueError("Must specify 'irttype' in irts!")
        else:
            irts['irttype']="none"

    if 'rcutoff' not in irtkeys:
        irts['rcutoff']=20

    if irts['irttype'] == "gamma":
        if 'gamma_beta' not in irtkeys:
            irts['gamma_beta'] = (1/1.1)
            #warnings.warn("Using default beta (Gamma IRT) weight of "+str(irts['gamma_beta']))
    if irts['irttype'] == "exgauss":
        if 'exgauss_lambda' not in irtkeys:
            irts['exgauss_lambda'] = 0.5
            #warnings.warn("Using default exgauss_lambda (Ex-Gaussian IRT) weight of "+str(irts['exgauss_lambda']))
        if 'exgauss_sigma' not in irtkeys:
            irts['exgauss_sigma'] = 0.5
            #warnings.warn("Using default exgauss_sigma (Ex-Gaussian IRT) weight of "+str(irts['exgauss_sigma']))

    return dotdict(irts)

def Fitinfo(fitinfo):
    fitkeys=list(fitinfo.keys())

    # full factorial of any list params
    for i in fitkeys:
        if isinstance(fitinfo[i],list):
            return flatten_list([Fitinfo(dict(fitinfo, **{i: j})) for j in fitinfo[i]])
    
    if 'followtype' not in fitkeys:
        fitinfo['followtype'] = "avg"   # or max or random
    if 'prior_method' not in fitkeys:
        fitinfo['prior_method'] = "zeroinflatedbetabinomial"
    if 'zibb_p' not in fitkeys:
        fitinfo['zibb_p'] = 0.5
    if 'prior_b' not in fitkeys:
        fitinfo['prior_b'] = 1
    if 'prior_a' not in fitkeys:        # adjust default prior_a depending on BB or ZIBB, to make edge prior .5
        if fitinfo['prior_method'] == "betabinomial":
            fitinfo['prior_a'] = fitinfo['prior_b']
        if fitinfo['prior_method'] == "zeroinflatedbetabinomial":
            fitinfo['prior_a'] = fitinfo['prior_b'] / float(fitinfo['zibb_p'])
    if 'directed' not in fitkeys:
        fitinfo['directed'] = False
    if 'startGraph' not in fitkeys:
        fitinfo['startGraph'] = "cn_valid"
    if 'prune_limit' not in fitkeys:
        fitinfo['prune_limit'] = np.inf
    if 'triangle_limit' not in fitkeys:
        fitinfo['triangle_limit'] = np.inf
    if 'other_limit' not in fitkeys:
        fitinfo['other_limit'] = np.inf
    if 'cn_windowsize' not in fitkeys:
        fitinfo['cn_windowsize'] = 2
    if 'cn_threshold' not in fitkeys:
        fitinfo['cn_threshold'] = 2
    if 'cn_alpha' not in fitkeys:
        fitinfo['cn_alpha'] = 0.05

    return dotdict(fitinfo)




# IO functions:
#
# * Read graph from file
# * Write graph to file
# * Read fluency data from file

from . import *

# wrapper
def graphToHash(a):
    return nx.generate_sparse6(nx.to_networkx_graph(a),header=False)

# wrapper
def hashToGraph(graphhash):
    return nx.to_numpy_matrix(nx.parse_graph6(graphhash))

# reads in graph from CSV
# row order not preserved; could be optimized more
def read_graph(fh,cols=(0,1),header=False,filters={},undirected=True,sparse=False):
    fh=open(fh,'rt', encoding='utf-8-sig')
    idx=0
    bigdict={}

    if header:
        headerrow=fh.readline().split('\n')[0].split(',')
        cols=(headerrow.index(cols[0]),headerrow.index(cols[1]))
        filterrows={}
        for i in list(filters.keys()):
            filterrows[headerrow.index(i)]=filters[i]
    else:
        filterrows={}

    done=dict()
    for line in fh:
        line=line.rstrip()
        linesplit=line.split(',')
        twoitems=[linesplit[cols[0]],linesplit[cols[1]]]
       
        skiprow=0
        for i in filterrows:
            if linesplit[i]!=filterrows[i]:
                skiprow=1
        if skiprow==1:
            continue
        
        try:
            if twoitems[1] not in bigdict[twoitems[0]]:
                bigdict[twoitems[0]].append(twoitems[1])
        except:
            bigdict[twoitems[0]] = [twoitems[1]]
        if twoitems[1] not in bigdict:              # doesn't this scale with dictionary size-- something i was trying to avoid by rewriting this function?
            bigdict[twoitems[1]] = []


    
    items_rev = dict(list(zip(list(bigdict.keys()),list(range(len(list(bigdict.keys())))))))
    items = dict(list(zip(list(range(len(list(bigdict.keys())))),list(bigdict.keys()))))
    
    if sparse:
        from scipy.sparse import csr_matrix
        rows=[]
        cols=[]
        numedges=0
        for i in bigdict:
            for j in bigdict[i]:
                rows.append(items_rev[i])
                cols.append(items_rev[j])
                numedges += 1
                if undirected:
                    rows.append(items_rev[j])
                    cols.append(items_rev[i])
                    numedges += 1
        data=np.array([1]*numedges)
        rows=np.array(rows)
        cols=np.array(cols)
        graph = csr_matrix((data, (rows, cols)), shape=(len(items),len(items)))
    else:
        graph = np.zeros((len(items),len(items)))
        
        for item1 in bigdict:
            for item2 in bigdict[item1]:
                idx1=items_rev[item1]
                idx2=items_rev[item2]
                graph[idx1,idx2]=1
                if undirected:
                    graph[idx2,idx1]=1

    return graph, items

# deprecated function name
def readX(*args, **kwargs):
    return load_fluency_data(*args, **kwargs)

# read Xs in from user files
# this should be re-written with pandas or something more managable
def load_fluency_data(filepath,category=None,removePerseverations=False,removeIntrusions=False,spell=None,scheme=None,group=None,subject=None,cleanBadChars=False,hierarchical=False):
   
    if type(group) is str:
        group = [group]
    if type(subject) is str:
        subject = [subject]
    if type(category) is str:
        category = [category]

    # grab header col indices
    mycsv = csv.reader(open(filepath,'rt', encoding='utf-8-sig'))
    headers = next(mycsv, None)
    subj_col = headers.index('id')
    listnum_col = headers.index('listnum')
    item_col = headers.index('item')
    
    # check for optional columns
    try:
        category_col = headers.index('category')
        has_category_col = True
    except:
        has_category_col = False
    try:
        group_col = headers.index('group')
        has_group_col = True
    except:
        has_group_col = False
        if group:
            raise ValueError('Data file does not have grouping column, but you asked for a specific group.')
    try:
        rt_col = headers.index('rt')
        has_rt_col = True
    except:
        has_rt_col = False

    Xs=dict()
    irts=dict()
    items=dict()
    spellingdict=dict()
    validitems=[]
    
    # read in list of valid items when removeIntrusions = True
    if removeIntrusions:
        if not scheme:
            raise ValueError('You need to provide a category scheme if you want to ignore intrusions!')
        else:
            with open(scheme,'rt', encoding='utf-8-sig') as fh:
                for line in fh:
                    if line[0] == "#": continue         # skip commented lines
                    try:
                        validitems.append(line.rstrip().split(',')[1].lower())
                    except:
                        pass    # fail silently on wrong format

    # read in spelling correction dictionary when spell is specified
    if spell:
        with open(spell,'rt', encoding='utf-8-sig') as spellfile:
            for line in spellfile:
                if line[0] == "#": continue         # skip commented lines
                try:
                    correct, incorrect = line.rstrip().split(',')
                    spellingdict[incorrect] = correct
                except:
                    pass    # fail silently on wrong format
   
    with open(filepath,'rt', encoding='utf-8-sig') as f:
        f.readline()    # discard header row
        for line in f:
            if line[0] == "#": continue         # skip commented lines
            row = line.rstrip().split(',')

            storerow = True  # if the row meets the filters specified then load it, else skip it
            if (subject != None) and (row[subj_col] not in subject):
                storerow = False
            if (group != None) and (row[group_col] not in group):
                storerow = False
            if (category != None) and (row[category_col] not in category):
                storerow = False
            
            if storerow == True:

                idx = row[subj_col]
                listnum_int = int(row[listnum_col])
                
                # make sure dict keys exist
                if idx not in Xs:
                    Xs[idx] = dict()
                    if has_rt_col:
                        irts[idx] = dict()
                if listnum_int not in Xs[idx]:
                    Xs[idx][listnum_int] = []
                    if has_rt_col:
                        irts[idx][listnum_int] = []
                if idx not in items:
                    items[idx] = dict()
                    
                # basic clean-up
                item=row[item_col].lower()
                if cleanBadChars:
                    badchars=" '-\"\\;?"
                    for char in badchars:
                        item=item.replace(char,"")
                if item in list(spellingdict.keys()):
                    item = spellingdict[item]
                if has_rt_col:
                    irt=row[rt_col]
                if item not in list(items[idx].values()):
                    if (item in validitems) or (not removeIntrusions):
                        item_count = len(items[idx])
                        items[idx][item_count]=item

                # add item to list
                try:
                    itemval=list(items[idx].values()).index(item)
                    if (not removePerseverations) or (itemval not in Xs[idx][listnum_int]):   # ignore any duplicates in same list resulting from spelling corrections
                        if (item in validitems) or (not removeIntrusions):
                            Xs[idx][listnum_int].append(itemval)
                            if has_rt_col:
                                irts[idx][listnum_int].append(int(irt)/1000.0)
                except:
                    pass                # bad practice to have empty except
   
    return Data({'Xs': Xs, 'items': items, 'irts': irts, 'structure': hierarchical})

def write_graph(gs, fh, subj="NA", directed=False, extra_data={}, header=True, labels=None, sparse=False):
    onezero={True: '1', False: '0'}        
    import networkx as nx
    fh=open(fh,'w')
    
    # if gs is not a list of graphs, then it should be a single graph
    if not isinstance(gs, list):
        gs = [gs]

    # turn them all into networkx graphs if they aren't already
    gs = [g if type(g) == nx.classes.graph.Graph else nx.to_networkx_graph(g) for g in gs]

    # label nodes if labels are provided
    if labels != None:
        if not isinstance(labels, list):
            labels = [labels]
        gs = [nx.relabel_nodes(i[0], i[1], copy=False) for i in zip(gs, labels)]
    
    nodes = list(set(flatten_list([list(gs[i].nodes()) for i in range(len(gs))])))
    
    if header == True:
        fh.write('subj,item1,item2,edge,\n')            # default header
    elif type(header) is str:
        fh.write('subj,item1,item2,'+ header+'\n')      # manual header if string is specified
    
    for node1 in nodes:
        for node2 in nodes:
            if (node1 < node2) or ((directed) and (node1 != node2)):   # write edges in alphabetical order unless directed graph
                edge = (node1,node2)
                edgelist=""
                if sparse:
                    write_edge = 0
                else:
                    write_edge = 1
                for g in gs:
                    hasedge = onezero[g.has_edge(edge[0],edge[1])]
                    edgelist=edgelist+"," + hasedge    # assumes graph is symmetrical if directed=True !!
                    write_edge += int(hasedge)
                
                if write_edge > 0:
                    extrainfo=""
                    if edge[0] in list(extra_data.keys()):
                        if edge[1] in list(extra_data[edge[0]].keys()):
                            if isinstance(extra_data[edge[0]][edge[1]],list):
                                extrainfo=","+",".join([str(i) for i in extra_data[sortededge[0]][sortededge[1]]])
                            else:
                                extrainfo=","+str(extra_data[sortededge[0]][sortededge[1]])
                    fh.write(subj    + "," +
                            str(edge[0])  + "," +
                            str(edge[1])  + 
                            edgelist + "," +
                            extrainfo + "\n")
    return



# This file is used to implement the 95% CI Clopper-Pearson method used by Community Network method.
# Originally, it relied on statsmodels/scipy, which is a very large package
from . import *

# copied from https://malishoaib.wordpress.com/2014/04/15/the-beautiful-beta-functions-in-raw-python/
def contfractbeta(a,b,x, ITMAX = 200):
    """ contfractbeta() evaluates the continued fraction form of the incomplete Beta function; incompbeta().  
    (Code translated from: Numerical Recipes in C.)"""
    EPS = 3.0e-7
    bm = az = am = 1.0
    qab = a+b
    qap = a+1.0
    qam = a-1.0
    bz = 1.0-qab*x/qap
    for i in range(ITMAX+1):
        em = float(i+1)
        tem = em + em
        d = em*(b-em)*x/((qam+tem)*(a+tem))
        ap = az + d*am
        bp = bz+d*bm
        d = -(a+em)*(qab+em)*x/((qap+tem)*(a+tem))
        app = ap+d*az
        bpp = bp+d*bz
        aold = az
        am = ap/bpp
        bm = bp/bpp
        az = app/bpp
        bz = 1.0
        if (abs(az-aold)<(EPS*abs(az))):
            return az
    print ('a or b too large or given ITMAX too small for computing incomplete beta function.')

# copied from https://malishoaib.wordpress.com/2014/04/15/the-beautiful-beta-functions-in-raw-python/
# same as scipy.special.betainc within rounding
# normalized incomplete beta is same as beta cdf
def incomplete_beta(a, b, x):
    ''' incompbeta(a,b,x) evaluates incomplete beta function, here a, b > 0 and 0 <= x <= 1. This function requires contfractbeta(a,b,x, ITMAX = 200) 
    (Code translated from: Numerical Recipes in C.)'''
    if (x == 0):
        return 0;
    elif (x == 1):
        return 1;
    else:
        lbeta = math.lgamma(a+b) - math.lgamma(a) - math.lgamma(b) + a * math.log(x) + b * math.log(1-x)
        if (x < (a+1) / (a+b+2)):
            return math.exp(lbeta) * contfractbeta(a, b, x) / a;
        else:
            return 1 - math.exp(lbeta) * contfractbeta(b, a, 1-x) / b;

# implements beta ppf
# same result as stats.beta.ppf(alpha_2, a, b)
def ppf(alpha_2, a, b, lower=0.0, upper=1.0, span=11, maxiter=20):
    if alpha_2 == 1.0:
        return 1.0
    elif alpha_2 == 0.0:
        return 0.0
    nprange = np.linspace(lower, upper, span)
    highlow = [incomplete_beta(a, b, x) > alpha_2 for x in nprange]
    idx_of_true = [idx for idx, x in enumerate(highlow) if x == True]
    if len(idx_of_true) == span:
        return lower
    elif len(idx_of_true) == 0:
        return upper
    else:
        if maxiter == 0:
            return nprange[idx_of_true[0]]
        else:
            return ppf(alpha_2, a, b, lower=nprange[idx_of_true[0]-1], upper=nprange[idx_of_true[0]], maxiter=(maxiter-1))

# same result as stats.beta.ppf(alpha_2, count, nobs - count + 1) (from statsmodels)
def pci_lowerbound(cooccur, total, alpha):
    alpha_2 = alpha * 0.5
    return ppf(alpha_2, cooccur, total - cooccur + 1)



from . import *

# frequency of each item in the data
# count also use collections.Counter ?
def freq(Xs, perlist=0):
    if perlist==1:
        Xs=[list(set(x)) for x in Xs]   # only count each item once per list
    Xflat=flatten_list(Xs)
    counts=[Xflat.count(i) for i in set(Xflat)]
    return dict(list(zip(set(Xflat),counts)))

# distribution of frequencies in the data (e.g., X items appeared once, Y items appeared twice, etc.)
def freq_stat(Xs):
    freqdist=list(freq(Xs).values())
    counts=[freqdist.count(i) for i in set(freqdist)]
    return dict(list(zip(set(freqdist),counts)))

# ** UNTESTED // modified from usf_cogsci.py
# num unique animals named across lists
def numNgrams(data, ngram=1):
    animallists=[find_ngrams(data[i],ngram) for i in range(len(data))]
    animalsNamed=len(set(flatten_list(animallists)))
    return animalsNamed



from . import *
from functools import reduce

# TODO: when doing same phase twice in a row, don't re-try same failures
    # (pass dict of failures, don't try if numchanges==0)
# TODO: Implement GOTM/ECN from Goni et al. 2011

def pf(*args, **kwargs):
    return pathfinder(*args, **kwargs)

def chan(*args, **kwargs):
    return pathfinder(*args, **kwargs)

def trimX(*args, **kwargs):
    return trim_lists(*args, **kwargs)

def genX(*args, **kwargs):
    return gen_lists(*args, **kwargs)

def goni(*args, **kwargs):
    return communityNetwork(*args, **kwargs)

def kenett(*args, **kwargs):
    return cbn(*args, **kwargs)
       
def correlationBasedNetwork(*args, **kwargs):
    return kenett(*args, **kwargs)

def nrw(*args, **kwargs):
    return naiveRandomWalk(*args, **kwargs)

# mix U-INVITE with random jumping model
def addJumps(probs, td, numnodes=None, statdist=None, Xs=None):
    """
    This is a test docstring
    """
    if (td.jumptype == "uniform") and (numnodes == None):
        raise ValueError("Must specify 'numnodes' when jumptype is uniform [addJumps]")
    if (td.jumptype == "stationary") and (np.any(statdist == None) or (Xs == None)):
        raise ValueError("Must specify 'statdist' and 'Xs' when jumptype is stationary [addJumps]")

    if td.jumptype == "uniform":
        jumpprob = float(td.jump)/numnodes                     # uniform jumping
    
    for l in range(len(probs)):                                # loop through all lists (l)
        for inum, i in enumerate(probs[l][1:]):                # loop through all items (i) excluding first (don't jump to item 1)
            if td.jumptype=="stationary":
                jumpprob = statdist[Xs[l][inum+1]]             # stationary probability jumping
                if np.isnan(jumpprob):                         # if node is disconnected, jumpprob is nan
                    jumpprob = 0.0
            probs[l][inum+1] = jumpprob + (1-td.jump)*i        # else normalize existing probability and add jumping probability
    return probs

# mix U-INVITE with priming model
# code is confusing...
def adjustPriming(probs, td, Xs):
    for xnum, x in enumerate(Xs[1:]):         # check all items starting with 2nd list
        for inum, i in enumerate(x[:-1]):     # except last item
            if i in Xs[xnum][:-1]:            # is item in previous list? if so, prime next item
                # follow prime with calc_prob_adjacent td.priming, follow RW with calc_prob_adjacent (1-td.priming)
                idx=Xs[xnum].index(i) # index of item in previous list
                if Xs[xnum][idx+1]==Xs[xnum+1][inum+1]:
                    probs[xnum+1][inum+1] = (probs[xnum+1][inum+1] * (1-td.priming)) + td.priming
                else:
                    probs[xnum+1][inum+1] = (probs[xnum+1][inum+1] * (1-td.priming))
    return probs

# Chan, Butters, Paulsen, Salmon, Swenson, & Maloney (1993) jcogneuro (p. 256) + pathfinder (PF; Schvaneveldt)
# + Chan, Butters, Salmon, Johnson, Paulsen, & Swenson (1995) neuropsychology
# Returns only PF(q, r) = PF(n-1, inf) = minimum spanning tree (sparsest possible graph)
# other parameterizations of PF(q, r) not implemented
def pathfinder(Xs, numnodes=None, valid=False, td=None):
    if numnodes == None:
        numnodes = len(set(flatten_list(Xs)))
    
    # From https://github.com/evanmiltenburg/dm-graphs
    def MST_pathfinder(G):
        """The MST-pathfinder algorithm (Quirin et al. 2008) reduces the graph to the
        unions of all minimal spanning trees."""
        NG    = nx.Graph()
        NG.add_nodes_from(list(range(numnodes)))
        edges = sorted( ((G[a][b]['weight'],a,b) for a,b in G.edges()),
                            reverse=False)                                                # smaller distances are more similar
        clusters = {node:i for i,node in enumerate(G.nodes())}
        while not edges == []:
            w1,a,b = edges[0]
            l      = []
            # Select edges to be considered this round:
            for w2,u,v in edges:
                if w1 == w2:
                    l.append((u,v,w2))
                else:
                    break
            # Remaining edges are those not being considered this round:
            edges = edges[len(l):]
            # Only select those edges for which the items are not in the same cluster
            l = [(a,b,c) for a,b,c in l if not clusters[a]==clusters[b]]
            # Add these edges to the graph:
            NG.add_weighted_edges_from(l)
            # Merge the clusters:
            for a,b,w in l:
                cluster_1 = clusters[a]
                cluster_2 = clusters[b]
                clusters = {node:cluster_1 if i==cluster_2 else i
                            for node,i in clusters.items()}
        return NG

    if valid and not td:
        raise ValueError('Need to pass DataModel when generating \'valid\' pathfinder()')
        
    N = float(len(Xs))
    distance_mat = np.zeros((numnodes, numnodes))
    for item1 in range(numnodes):
        for item2 in range(item1+1,numnodes):
            Tij = 0
            dijk = 0
            for x in Xs:
                if (item1 in x) and (item2 in x):
                    Tij += 1
                    dijk = dijk + (abs(x.index(item1) - x.index(item2)) / float(len(x)))
            try:
                dij = dijk * (N / (Tij**2))
            except:
                dij = 0.0      # added constraint for divide-by-zero... this will ensure that no edge will exist between i and j
            distance_mat[item1, item2] = dij
            distance_mat[item2, item1] = dij

    #graph = scipy.sparse.csgraph.minimum_spanning_tree(distance_mat)
    graph = nx.to_numpy_matrix(MST_pathfinder(nx.Graph(distance_mat)))

    # binarize and make graph symmetric (undirected)... some redundancy but it's cheap
    #graph = np.where(graph.todense(), 1, 0)
    graph = np.array(np.where(graph, 1, 0))
    for rownum, row in enumerate(graph):
        for colnum, val in enumerate(row):
            if val==1:
                graph[rownum,colnum]=1
                graph[colnum,rownum]=1

    if valid:
        graph = makeValid(Xs, graph, td)
        
    #return np.array(graph).astype(int)
    return graph

# objective graph cost
# returns the number of links that need to be added or removed to reach the true graph
def cost(graph,a, undirected=True):
    cost = sum(sum(np.array(abs(graph-a))))
    if undirected:
        return cost/2.0
    else:
        return cost

# graph=estimated graph, a=target/comparison graph
def costSDT(graph, a):
    hit=0; miss=0; fa=0; cr=0
    check=(graph==a)
    for rnum, r in enumerate(a):
        for cnum, c in enumerate(r[:rnum]):
            if check[rnum,cnum]==True:
                if a[rnum,cnum]==1:
                    hit += 1
                else:
                    cr += 1
            else:
                if a[rnum,cnum]==1:
                    miss += 1
                else:
                    fa += 1
    return [hit, miss, fa, cr]

def evalGraphPrior(a, prior, undirected=True):
    probs = []
    priordict = prior[0]
    items = prior[1]
    nullprob = priordict['DEFAULTPRIOR']

    for inum, i in enumerate(a):
        for jnum, j in enumerate(i):
            if (inum > jnum) or ((undirected==False) and (inum != jnum)):
                if undirected:
                    pair = np.sort((items[inum], items[jnum]))
                else:
                    pair = (items[inum], items[jnum])
                try:
                    priorprob = priordict[pair[0]][pair[1]]
                    if j==1:
                        prob = priorprob
                    elif j==0:
                        prob = (1-priorprob)
                except:
                    prob = nullprob  #  no information about edge
                probs.append(prob)
    
    probs = [np.log(prob) for prob in probs]      # multiplication probably results in underflow...
    probs = sum(probs)
    return probs

def firstEdge(Xs, numnodes=None):
    if numnodes == None:
        numnodes = len(set(flatten_list(Xs)))
    a=np.zeros((numnodes,numnodes))
    for x in Xs:
        a[x[0],x[1]]=1
        a[x[1],x[0]]=1 # symmetry
    a=np.array(a.astype(int))
    return a

def fullyConnected(numnodes):
    a=np.ones((numnodes,numnodes))
    for i in range(numnodes):
        a[i,i]=0.0
    return a.astype(int)

# only returns adjacency matrix, not nx graph
def naiveRandomWalk(Xs, numnodes=None, directed=False):
    if numnodes == None:
        numnodes = len(set(flatten_list(Xs)))
    a=np.zeros((numnodes,numnodes))
    for x in Xs:
        walk = edges_from_nodes(x)
        for i in set(walk):
            a[i[0],i[1]]=1
            if directed==False:
                a[i[1],i[0]]=1
    a=np.array(a.astype(int))
    return a

def genGraphPrior(graphs, items, fitinfo=Fitinfo({}), mincount=1, undirected=True, returncounts=False):
    a_start = fitinfo.prior_a
    b_start = fitinfo.prior_b
    method = fitinfo.prior_method
    p = fitinfo.zibb_p
    
    priordict={}
  
    def betabinomial(a,b):
        return (b / (a + b))
    
    def zeroinflatedbetabinomial(a,b,p):
        return (b / ((1-p)*a+b))

    # tabulate number of times edge does or doesn't appear in all of the graphs when node pair is present
    for graphnum, graph in enumerate(graphs):   # for each graph
        itemdict=items[graphnum]
        for inum, i in enumerate(graph):     # rows of graph
            for jnum, j in enumerate(i):     # columns of graph
                if (inum > jnum) or ((undirected==False) and (inum != jnum)):
                    item1 = itemdict[inum]
                    item2 = itemdict[jnum]
                    if undirected:
                        pair = np.sort((item1,item2))
                    else:
                        pair = (item1,item2)
                    if pair[0] not in list(priordict.keys()):
                        priordict[pair[0]]={}
                    if pair[1] not in list(priordict[pair[0]].keys()):
                        priordict[pair[0]][pair[1]] = [a_start, b_start]
                    if j==1:
                        priordict[pair[0]][pair[1]][1] += 1.0 # increment b counts
                    elif j==0:
                        priordict[pair[0]][pair[1]][0] += 1.0 # increment a counts
   
    if not returncounts:
        for item1 in priordict:
            for item2 in priordict[item1]:
                a, b = priordict[item1][item2]      # a=number of participants without link, b=number of participants with link
                if (a+b) >= (mincount + a_start + b_start):
                    if method == "zeroinflatedbetabinomial":
                        priordict[item1][item2] = zeroinflatedbetabinomial(a,b,p) # zero-inflated beta-binomial
                    elif method == "betabinomial":
                        priordict[item1][item2] = betabinomial(a,b) # beta-binomial
                else:
                    priordict[item1][item2] = 0.0   # if number of observations is less than mincount, make edge prior 0.0
        if 'DEFAULTPRIOR' in list(priordict.keys()):
            raise ValueError('Sorry, you can\'t have a node called DEFAULTPRIOR. \
                              Sure, I should have coded this better, but I really didn\'t think this situation would ever occur.')
        else:
            if method == "zeroinflatedbetabinomial":
                priordict['DEFAULTPRIOR'] = zeroinflatedbetabinomial(a_start, b_start, p)
            elif method=="betabinomial":
                priordict['DEFAULTPRIOR'] = betabinomial(a_start, b_start)
    
    return priordict

# generate starting graph for U-INVITE
def genStartGraph(Xs, numnodes, td, fitinfo):
    if fitinfo.startGraph=="cn_valid":
        graph = communityNetwork(Xs, numnodes, td=td, valid=True, fitinfo=fitinfo)
    elif fitinfo.startGraph=="pf_valid":
        graph = pathfinder(Xs, numnodes, valid=True, td=td)
    elif (fitinfo.startGraph=="rw" or fitinfo.startGraph=="nrw"):
        graph = naiveRandomWalk(Xs,numnodes)
    elif fitinfo.startGraph=="fully_connected":
        graph = fullyConnected(numnodes)
    elif fitinfo.startGraph=="empty_graph":
        graph = np.zeros((numnodes,numnodes)).astype(int)           # useless...
    else:
        graph = np.copy(fitinfo.startGraph)                         # assume a graph has been passed as a starting point
    return graph

# w = window size; two items appear within +/- w steps of each other (where w=1 means adjacent items)
# f = filter frequency; if two items don't fall within the same window more than f times, then no edge is inferred
# c = confidence interval; retain the edge if there is a <= c probability that two items occur within the same window n times by chance alone
# valid (t/f) ensures that graph can produce data using censored RW.
def communityNetwork(Xs, numnodes=None, fitinfo=Fitinfo({}), valid=False, td=None):
    if numnodes == None:
        numnodes = len(set(flatten_list(Xs)))
        
    w = fitinfo.cn_windowsize
    f = fitinfo.cn_threshold
    c = fitinfo.cn_alpha
    
    if f<1:                 # if <1 treat as proportion of total lists; if >1 treat as absolute # of lists
        f=int(round(len(Xs)*f))

    if valid and not td:
        raise ValueError('Need to pass Data when generating \'valid\' goni()')

    #if c<1:
    #    from statsmodels.stats.proportion import proportion_confint as pci

    if w < 1:
        print("Error in goni(): w must be >= 1")
        return

    graph=np.zeros((numnodes, numnodes)).astype(int)         # empty graph

    # frequency of co-occurrences within window (w)
    for x in Xs:                                             # for each list
        for pos in range(len(x)):                            # for each item in list
            for i in range(1, w+1):                          # for each window size
                if pos+i<len(x):
                    graph[x[pos],x[pos+i]] += 1
                    graph[x[pos+i],x[pos]] += 1

    # exclude edges with co-occurrences less than frequency (f) and binarize
    # but first save co-occurence frequencies
    cooccur = np.copy(graph)
    for i in range(len(graph)):
        for j in range(len(graph)):
            if graph[i, j] < f:
                graph[i, j] = 0
            else:
                graph[i, j] = 1

    # check if co-occurrences are due to chance
    if c<1:
        setXs=[list(set(x)) for x in Xs]                              # unique nodes in each list
        flatX=flatten_list(setXs)                                     # flattened
        xfreq=[flatX.count(i) for i in range(numnodes)]               # number of lists each item appears in (at least once)
        listofedges=list(zip(*np.nonzero(graph)))                           # list of edges in graph to check
        numlists=float(len(Xs))
        meanlistlength=np.mean([len(x) for x in Xs])
    
        # Goni et al. (2011), eq. 10
        p_adj = (2.0/(meanlistlength*(meanlistlength-1))) * ((w*meanlistlength) - ((w*(w+1))/2.0))
        for i,j in listofedges:
            p_linked = (xfreq[i]/numlists) * (xfreq[j]/numlists) * p_adj
            #ci=pci(cooccur[i,j],numlists,alpha=c,method="beta")[0]    # lower bound of Clopper-Pearson binomial CI
            ci = pci_lowerbound(cooccur[i,j], numlists, c)             # lower bound of Clopper-Pearson binomial CI
            if p_linked >= ci:                                        # if co-occurrence could be due to chance, remove edge
                graph[i,j]=0
                graph[j,i]=0

    if valid:
        graph = makeValid(Xs, graph, td)

    # make sure there are no self-transitions -- is this necessary?
    for i in range(len(graph)):
        graph[i,i]=0.0

    return graph

# enrich graph by finding modules and making them completely interconnected
# using Generalized Topological Overlap Measure (GTOM)
# right now only does n=2 (neighbors and neighbors of neighbors)
# see Goni et al 2010
# TODO unfinished: so far, creates GTOM matrix but doesn't "enrich" network... how to determine # of clusters?
def gtom(graph):

    # modified from uinvite(), copied for convenience (TODO consolidate by moving outside to its own function)
    # return list of neighbors of neighbors of i, that aren't themselves neighbors of i
    # i.e., an edge between i and any item in nn forms a triangle
    def neighborsofneighbors(i, nxg):
        nn=[]                                       # neighbors of neighbors (nn)
        n=list(nx.all_neighbors(nxg,i))
        for j in n:
            nn=nn+list(nx.all_neighbors(nxg,j))
        nn=list(set(nn))
        if i in nn:
            nn.remove(i)                            # remove self
        return nn
    
    nxgraph = nx.to_networkx_graph(graph)
    numnodes = nx.number_of_nodes(nxgraph)
    gtom_mat = np.zeros((numnodes,numnodes))
   
    nn_dict = {}
    for i in range(numnodes):
        nn_dict[i] = neighborsofneighbors(i, nxgraph)
    
    for i in range(numnodes):
        for j in range(i+1,numnodes):
            i_neighbors = nn_dict[i]
            j_neighbors = nn_dict[j]
            min_neighbors = min(len(i_neighbors),len(j_neighbors))
            len_overlap = len(set.intersection(set(i_neighbors),set(j_neighbors)))
            gtom_mat[i, j] = 1 - (float(len_overlap) / min_neighbors)
            gtom_mat[j, i] = gtom_mat[i, j]

    return gtom_mat

 
def hierarchicalUinvite(Xs, items, numnodes, td, irts=False, fitinfo=Fitinfo({}), seed=None, debug=True):
    nplocal=np.random.RandomState(seed) 
    fitinfoSG = fitinfo.startGraph  # fitinfo is mutable, need to revert at end of function... blah
    # create ids for all subjects
    subs=list(range(len(Xs)))
    graphs=[[]]*len(subs)

    # cycle though participants
    exclude_subs=[]
    graphchanges=1
    rnd=1
    while graphchanges > 0:
        if debug: print("Round: ", rnd)
        graphchanges = 0
        nplocal.shuffle(subs)
        for sub in [i for i in subs if i not in exclude_subs]:
            if debug: print("SS: ", sub)
            
            #td.numx = len(Xs[sub])
            if graphs[sub] == []:
                fitinfo.startGraph = fitinfoSG      # on first pass for subject, use default fitting method (e.g., NRW, goni, etc)
            else:
                fitinfo.startGraph = graphs[sub]    # on subsequent passes, use ss graph from previous iteration

            # generate prior without participant's data, fit graph
            priordict = genGraphPrior(graphs[:sub]+graphs[sub+1:], items[:sub]+items[sub+1:], fitinfo=fitinfo)
            prior = (priordict, items[sub])
            
            if isinstance(irts, list):
                uinvite_graph, bestval = uinvite(Xs[sub], td, numnodes[sub], fitinfo=fitinfo, prior=prior, irts=irts[sub])
            else:
                uinvite_graph, bestval = uinvite(Xs[sub], td, numnodes[sub], fitinfo=fitinfo, prior=prior)

            if not np.array_equal(uinvite_graph, graphs[sub]):
                graphchanges += 1
                graphs[sub] = uinvite_graph
                exclude_subs=[sub]              # if a single change, fit everyone again (except the graph that was just fit)
            else:
                exclude_subs.append(sub)        # if graph didn't change, don't fit them again until another change
        rnd += 1

    ## generate group graph
    priordict = genGraphPrior(graphs, items, fitinfo=fitinfo)
    fitinfo.startGraph = fitinfoSG  # reset fitinfo starting graph to default
    
    return graphs, priordict

def probXhierarchical(Xs, graphs, items, td, priordict=None, irts=Irts({})):
    lls=[]
    for sub in range(len(Xs)):
        if priordict:
            prior = (priordict, items[sub])
        else:
            prior=None
        best_ll, probmat = probX(Xs[sub], graphs[sub], td, irts=irts, prior=prior)   # LL of graph
        lls.append(best_ll)
    ll=sum(lls)
    return ll

# construct graph using method using item correlation matrix and planar maximally filtered graph (PMFG)
# see Borodkin, Kenett, Faust, & Mashal (2016) and Kenett, Kenett, Ben-Jacob, & Faust (2011)
# does not work well for small number of lists! many NaN correlations + when two correlations are equal, ordering is arbitrary
def correlationBasedNetwork(Xs, numnodes, minlists=0, valid=False, td=None):
    if valid and not td:
        raise ValueError('Need to pass Data when generating \'valid\' kenett()')
    
    import planarity
    
    # construct matrix of list x item where each cell indicates whether that item is in that list
    list_by_item = np.zeros((numnodes,len(Xs)))
    for node in range(numnodes):
        for x in range(len(Xs)):
            if node in Xs[x]:
                list_by_item[node,x]=1.0
    
    # find pearsonr correlation for all item pairs
    item_by_item = {}
    for item1 in range(numnodes):
        for item2 in range(item1+1,numnodes):
            if (sum(list_by_item[item1]) <= minlists) or (sum(list_by_item[item2]) <= minlists):    # if a word only appears in <= minlists lists, exclude from correlation list (kenett personal communication, to avoid spurious correlations)
                item_by_item[(item1, item2)] = np.nan
            else:
                #item_by_item[(item1, item2)] = scipy.stats.pearsonr(list_by_item[item1],list_by_item[item2])[0]
                item_by_item[(item1, item2)] = pearsonr(list_by_item[item1],list_by_item[item2])
    
    corr_vals = sorted(item_by_item, key=item_by_item.get)[::-1]       # keys in correlation dictionary sorted by value (high to low, including NaN first)

    edgelist=[]
    for pair in corr_vals:
        if not np.isnan(item_by_item[pair]):    # nan correlation occurs when item is in all lists-- exclude from graph (conservative)
            edgelist.append(pair)
            if not planarity.is_planar(edgelist):
                edgelist.pop()
    
    g = nx.Graph()
    g.add_nodes_from(list(range(numnodes)))
    g.add_edges_from(edgelist)
    a=np.array(nx.to_numpy_matrix(g)).astype(int)
   
    if valid:
        a = makeValid(Xs, a, td)
   
    return a

def makeValid(Xs, graph, td, seed=None):
    # add direct edges when transition is impossible
    check=probX(Xs, graph, td)
    while check[0] == -np.inf:
        if isinstance(check[1],tuple):
            graph[check[1][0], check[1][1]] = 1
            graph[check[1][1], check[1][0]] = 1
        # i think these 2 lines are no longer necessary
        #elif check[1] == "prior":
        #    raise ValueError('Starting graph has prior probability of 0.0')
        elif isinstance(check[1],int):
            # when list contains one item and node is unreachable, connect to random node
            nplocal = np.random.RandomState(seed)
            randnode = nplocal.choice(range(len(graph)))
            graph[check[1], randnode] = 1
            graph[randnode, check[1]] = 1
        else:
            raise ValueError('Unexpected error from makeValid()')
        check=probX(Xs, graph, td)
    return graph

# converts priordict to graph if probability of edge is greater than cutoff value
def priorToGraph(priordict, items, cutoff=0.5, undirected=True):
    numnodes = len(items)
    a = np.zeros((numnodes, numnodes))
    
    for item1 in list(priordict.keys()):
        if item1 != 'DEFAULTPRIOR':
            for item2 in priordict[item1]:
                if priordict[item1][item2] > cutoff:
                    item1_idx = list(items.keys())[list(items.values()).index(item1)]    # syntax is a little convoluted in case dictionary keys are not in sequential order
                    item2_idx = list(items.keys())[list(items.values()).index(item2)]
                    a[item1_idx, item2_idx] = 1.0
                    if undirected:
                        a[item2_idx, item1_idx] = 1.0
    return a

# probability of observing Xs, including irts and prior
#@profile
#@nogc
def probX(Xs, a, td, irts=Irts({}), prior=None, origmat=None, changed=[]):

    numnodes=len(a)
    reg=(1+1e-10)                           # nuisance parameter to prevent errors; can also use pinv instead of inv, but that's much slower
    identmat=np.identity(numnodes) * reg    # pre-compute for tiny speed-up (only for non-IRT)

    probs=[]

    # generate transition matrix (from: column, to: row) from link matrix
    t=a/sum(a.astype(float))
    t=np.nan_to_num(t)                  # jumping/priming models can have nan in matrix, need to change to 0
    
    if (td.jumptype=="stationary") or (td.start_node=="stationary"):
        statdist=stationary(t)

    # U-INVITE probability excluding jumps, prior, and priming adjustments -- those come later
    for xnum, x in enumerate(Xs):
        x2=np.array(x)
        t2=t[x2[:,None],x2]                                        # re-arrange transition matrix to be in list order
        prob=[]
        if td.start_node=="stationary":
            prob.append(statdist[x[0]])                            # probability of X_1
        elif td.start_node=="uniform":
            prob.append(1.0/numnodes)

        # if impossible starting point, return immediately
        if (prob[-1]==0.0):
            try:
                return -np.inf, (x[0], x[1])
            except:
                return -np.inf, x[0]

        if (len(changed) > 0) and isinstance(origmat,list):        # if updating prob. matrix based on specific link changes
            update=0                                               # reset for each list

        # flag if list contains perseverations
        if len(x) == len(set(x)):
            list_has_perseverations = False
        else:
            list_has_perseverations = True

        for curpos in range(1,len(x)):
            if (len(changed) > 0) and isinstance(origmat,list):
                if update==0:                                      # first check if probability needs to be updated
                    if (Xs[xnum][curpos-1] in changed):            # (only AFTER first changed node has been reached)
                        update=1
                    else:                                          # if not, take probability from old matrix
                        prob.append(origmat[xnum][curpos])
                        continue
            
            if list_has_perseverations:            # a bit slower because matrix is being copied
                x2=np.array([i for i,j in enumerate(x) if (j not in x[:i]) and (i < curpos)]) # column ids for transient states excluding perseverations
                Q=t2[x2[:,None],x2]                # excludes perseverations. could be sped if only performed when Q contains perseverations
                                                   # as opposed to being done for every transition if a perseveration is in the list
            else:                                  
                Q=t2[:curpos,:curpos]              # old way when data does not include perseverations
                
            # td.censor_fault is necessary to model perservations in the data
            if td.censor_fault > 0.0:
                Q=np.multiply(Q, 1.0-td.censor_fault)
            
            if len(irts.data) > 0:     # use this method only when passing IRTs
                numcols=len(Q)
                flist=[]
                newQ=np.zeros(numcols)                             # init to Q^0, for when r=1
                newQ[curpos-1]=1.0                                 # (using only one: row for efficiency)

                irt=irts.data[xnum][curpos-1]

                # precompute for small speedup
                if irts.irttype=="gamma":
                    logbeta=np.log(irts.gamma_beta)
                    logirt=np.log(irt)

                # normalize irt probabilities to avoid irt weighting
                if irts.irttype=="gamma":
                     # r=alpha. probability of observing irt at r steps
                    irtdist=[r*logbeta-math.lgamma(r)+(r-1)*logirt-irts.gamma_beta*irt for r in range(1,irts.rcutoff)]
                if irts.irttype=="exgauss":
                
                    irtdist=[np.log(irts.exgauss_lambda/2.0)+(irts.exgauss_lambda/2.0)*(2.0*r+irts.exgauss_lambda*(irts.exgauss_sigma**2)-2*irt)+np.log(math.erfc((r+irts.exgauss_lambda*(irts.exgauss_sigma**2)-irt)/(np.sqrt(2)*irts.exgauss_sigma))) for r in range(1,irts.rcutoff)]

                for r in range(1,irts.rcutoff):
                    innersum=0
                    for k in range(numcols):
                        num1=newQ[k]                               # probability of being at node k in r-1 steps
                        num2=t2[curpos,k]                          # probability transitioning from k to absorbing node    
                        innersum=innersum+(num1*num2)

                    # compute irt probability given r steps
                    log_dist = irtdist[r-1] / sum(irtdist)

                    if innersum > 0: # sometimes it's not possible to get to the target node in r steps
                        flist.append(log_dist + np.log(innersum))

                    newQ=np.inner(newQ,Q)                          # raise power by one

                f=sum([np.e**i for i in flist])
                prob.append(f)                                     # probability of x_(t-1) to X_t
            else:                                                  # if no IRTs, use standard U-INVITE
                I=identmat[:len(Q),:len(Q)]
                
                # novel items are emitted with probability 1 when encountered. perseverations are emitted with probability td.censor_fault when encountered.
                if list_has_perseverations:              # if list has perseverations. could speed up by only doing this step when a perseveration has been encountered
                    x1=np.array([curpos])   # absorbing node
                    #x2=np.array([i for i,j in enumerate(x) if (j not in x[:i]) and (i < curpos)]) # column ids for transient states excluding perseverations
                    x2=np.array([i for i,j in enumerate(x) if (j not in x[i+1:curpos]) and (i < curpos)]) # column ids for transient states excluding perseverations
                    R=t2[x1[:,None],x2][0]  # why is [0] necessary here but not in the else case?
                    
                    if Xs[xnum][curpos] in Xs[xnum][:curpos]:       # if absorbing state has appeared in list before...
                        R=np.multiply(R,td.censor_fault)
                else:                                       # if not a perseveration
                    R=t2[curpos,:curpos]                    # old way
               
                ### test (when censor_fault=0) to see if absorbing distribution sums to 1... something is broken
                #total = []
                #x2=np.array([j for i,j in enumerate(x) if (i < curpos)]) # column ids for transient states excluding perseverations
                #N=np.linalg.solve(I-Q,I[-1])
                #for i in range(len(t)):
                #    R=t[np.array([i])[:,None],x2]
                #    B=np.dot(R,N)
                #    total.append(B[0])
                #    if B[0] > 1.0:
                #        print("NONONO")
                #print("total ", total)
                #R=t2[curpos,:curpos]                    # old way to reset
                ###
                
                N=np.linalg.solve(I-Q,I[-1])
                B=np.dot(R,N)
                if np.isnan(B):
                    B=0.0
                prob.append(B)
               
                # alternative/original using matrix inverse
                #R=t2[curpos:,:curpos]
                #N=inv(I-Q)
                #B=np.dot(R,N)                
                #prob.append(B[0,curpos-1])

            # if there's an impossible transition and no jumping/priming, return immediately
            if (prob[-1]==0.0) and (td.jump == 0.0) and (td.priming == 0.0):
                return -np.inf, (x[curpos-1], x[curpos])

        probs.append(prob)

    uinvite_probs = copy.deepcopy(probs)      # store only u-invite transition probabilities (the computationally hard stuff) to avoid recomputing
    
    # adjust for jumping probability
    if td.jump > 0.0:
        if td.jumptype=="uniform":
            probs=addJumps(probs, td, numnodes=numnodes)
        elif td.jumptype=="stationary":
            probs=addJumps(probs, td, statdist=statdist, Xs=Xs)

    if (td.priming > 0.0):
        probs=adjustPriming(probs, td, Xs)

    # check for impossible transitions after priming and jumping
    for xnum, x in enumerate(probs):
        for inum, i in enumerate(x):
            if (i==0.0) and (inum==0):
                return -np.inf, (Xs[xnum][inum], Xs[xnum][inum+1])  # link to next item when first item is unreachable
            elif (i==0.0) and (inum > 0):
                return -np.inf, (Xs[xnum][inum-1], Xs[xnum][inum])  # link to previous item otherwise
                
    try:
        ll=sum([sum([np.log(j) for j in probs[i]]) for i in range(len(probs))])
    except:
        ll=-np.inf

    # include prior?
    if prior:
        priorlogprob = evalGraphPrior(a, prior)
        ll = ll + priorlogprob

    return ll, uinvite_probs

#@profile
def uinvite(Xs, td, numnodes=None, irts=Irts({}), fitinfo=Fitinfo({}), prior=None, debug=True, seed=None):
    nplocal=np.random.RandomState(seed) 

    if numnodes == None:
        numnodes = len(set(flatten_list(Xs)))

    # return list of neighbors of neighbors of i, that aren't themselves neighbors of i
    # i.e., an edge between i and any item in nn forms a triangle
    #@profile
    def neighborsofneighbors(i, nxg):
        nn=[]                                       # neighbors of neighbors (nn)
        n=list(nx.all_neighbors(nxg,i))
        for j in n:
            nn=nn+list(nx.all_neighbors(nxg,j))
        nn=list(set(nn))
        for k in n:                                 # remove neighbors
            if k in nn:
                nn.remove(k)
        if i in nn:
            nn.remove(i)                            # remove self
        return nn
        
    # toggle links back, should be faster than making graph copy
    #@profile
    def swapEdges(graph,links):
        for link in links:
            graph[link[0],link[1]] = 1 - graph[link[0],link[1]]
            if not fitinfo.directed:
                graph[link[1],link[0]] = 1 - graph[link[1],link[0]]
        return graph
        
    #@timer
    #@profile
    def pivot(graph, vmin=1, vmaj=0, best_ll=None, probmat=None, limit=np.inf, method=""):
        numchanges=0     # number of changes in single pivot() call

        if (best_ll == None) or (np.any(probmat == None)):
            best_ll, probmat = probX(Xs,graph,td,irts=irts,prior=prior)   # LL of best graph found
        nxg=nx.to_networkx_graph(graph)

        # generate dict where v[i] is a list of nodes where (i, v[i]) is an existing edge in the graph
        if (method=="prune") or (method==0):
            if debug:
                print("Pruning", str(vmaj) + "." + str(vmin), "... ",) # (len(edges)/2)-len(firstedges), "possible:",
            sys.stdout.flush()
            listofedges=np.where(graph==1)
            v=dict()
            for i in range(numnodes):
                v[i]=[]
            for i in zip(*listofedges):
                if ((i[0], i[1]) not in firstedges) and ((i[1], i[0]) not in firstedges): # don't flip first edges (FE)!
                    if td.jump == 0.0:                                                      # unless jumping is allowed, untested 10/6/17 JCZ
                        v[i[0]].append(i[1])
        
        # generate dict where v[i] is a list of nodes where (i, v[i]) would form a new triangle
        if (method=="triangles") or (method==1):
            if debug:
                print("Adding triangles", str(vmaj) + "." + str(vmin), "... ",) # (len(edges)/2), "possible:",
            sys.stdout.flush()
            nn=dict()
            for i in range(len(graph)):
                nn[i]=neighborsofneighbors(i, nxg)
            v=nn
        
        # generate dict where v[i] is a list of nodes where (i, v[i]) is NOT an existing an edge and does NOT form a triangle
        if (method=="nonneighbors") or (method==2):
            # list of a node's non-neighbors (non-edges) that don't form triangles
            if debug:
                print("Adding other edges", str(vmaj) + "." + str(vmin), "... ",)
            sys.stdout.flush()
            nonneighbors=dict()
            for i in range(numnodes):
                nn=neighborsofneighbors(i, nxg)
                # non-neighbors that DON'T form triangles 
                nonneighbors[i]=[j for j in range(numnodes) if j not in nx.all_neighbors(nxg,i) and j not in nn] 
                nonneighbors[i].remove(i) # also remove self
            v=nonneighbors

        count=[0.0]*numnodes
        avg=[-np.inf]*numnodes
        finishednodes=0
        loopcount=0

        while (finishednodes < numnodes) and (loopcount < limit):
            loopcount += 1          # number of failures before giving up on this pahse
            maxval=max(avg)             
            bestnodes=[i for i, j in enumerate(avg) if j == maxval]  # most promising nodes based on avg logprob of edges with each node as vertex
            node1=nplocal.choice(bestnodes)

            if len(v[node1]) > 0:
                #node2=nplocal.choice(v[node1]) # old
                
                n2avg=[avg[i] for i in v[node1]]
                maxval=max(n2avg)
                bestnodes=[v[node1][i] for i, j in enumerate(n2avg) if j == maxval]
                node2=nplocal.choice(bestnodes)

                edge=(node1, node2)
                graph=swapEdges(graph,[edge])

                graph_ll, newprobmat=probX(Xs,graph,td,irts=irts,prior=prior,origmat=probmat,changed=[node1,node2])

                if best_ll >= graph_ll:
                    graph=swapEdges(graph,[edge])
                else:
                    best_ll = graph_ll
                    probmat = newprobmat
                    numchanges += 1
                    loopcount = 0
                    # probX under all possible perseveration values JCZ 5/9/2018
                    if fitinfo.estimatePerseveration:
                        old_censor = td.censor_fault
                        best_param = old_censor
                        for censor_param in [i/100.0 for i in range(101)]:
                            td.censor_fault = censor_param
                            graph_ll, newprobmat = probX(Xs,graph,td,irts=irts,prior=prior)   # LL of starting graph
                            if graph_ll > best_ll:
                                best_ll = graph_ll
                                probmat = newprobmat
                                best_param = censor_param
                        td.censor_fault = best_param
                        print("parameter old:", old_censor, " new: ", best_param)

                v[node1].remove(node2)   # remove edge from possible choices
                if not fitinfo.directed:
                    v[node2].remove(node1)
           
                # increment even if graph prob = -np.inf for implicit penalty
                count[node1] += 1
                count[node2] += 1
                if (graph_ll != -np.inf) and (fitinfo.followtype != "random"):
                    if avg[node1] == -np.inf:
                        avg[node1] = graph_ll
                    else: # followtype == avg
                            avg[node1] = avg[node1] * ((count[node1]-1)/count[node1]) + (1.0/count[node1]) * graph_ll
                    if avg[node2] == -np.inf:
                        avg[node2] = graph_ll
                    else:  # followtype == avg
                        avg[node2] = avg[node2] * ((count[node2]-1)/count[node2]) + (1.0/count[node2]) * graph_ll
            else:                       # no edges on this node left to try!
                avg[node1]=-np.inf      # so we don't try it again...
                finishednodes += 1

        if debug:
            print(numchanges, "changes")

        return graph, best_ll, probmat, numchanges

    #    return graph
    
    def phases(graph, best_ll, probmat):
        complete=[0,0,0]         # marks which phases are complete
        vmaj=0
        vmin=1
        while sum(complete) < 3:
            phasenum=complete.index(0)
            if phasenum==0: limit=fitinfo.prune_limit
            if phasenum==1: limit=fitinfo.triangle_limit
            if phasenum==2: limit=fitinfo.other_limit
            if (phasenum==0) and (vmin==1): vmaj += 1

            graph, best_ll, probmat, numchanges = pivot(graph, best_ll=best_ll, vmaj=vmaj, vmin=vmin, method=phasenum, limit=limit, probmat=probmat)
            if numchanges > 0:
                vmin += 1
            else:
                if (vmin==1): complete[phasenum]=1
                if (phasenum==0) and (vmin>1): complete=[1,0,0]
                if (phasenum==1) and (vmin>1): complete=[0,1,0]
                if (phasenum==2) and (vmin>1): complete=[0,0,1]
                vmin=1
        return graph, best_ll

    # check if data has perseverations
    if not [len(set(i)) for i in Xs]==[len(i) for i in Xs]:
        if (td.censor_fault == 0.0) and (not fitinfo.estimatePerseveration):
            raise Exception('''Your data contains perseverations, but \
                            censor_fault = 0.0; Set to some value 0.0 < x <= 1.0 or
                            set estimatePerseverations to True''')

    try:
        firstedges=[(x[0], x[1]) for x in Xs]
    except:
        firstedges=[]
    
    # find a good starting graph
    graph = genStartGraph(Xs, numnodes, td, fitinfo)

    # find best starting perseveration parameter if applicable JCZ 5/9/2018
    if fitinfo.estimatePerseveration:
        best_ll = -np.inf
        best_param = 0.0
        for censor_param in [i/100.0 for i in range(101)]:
            td.censor_fault = censor_param
            graph_ll, probmat = probX(Xs,graph,td,irts=irts,prior=prior)   # LL of starting graph
            if graph_ll > best_ll:
                best_ll = graph_ll
                best_param = censor_param
        td.censor_fault = best_param
        
    best_ll, probmat = probX(Xs,graph,td,irts=irts,prior=prior)   # LL of starting graph
    graph, best_ll = phases(graph, best_ll, probmat)

    return graph, best_ll



from . import *

def degree_dist(g):
    if isinstance(g,np.ndarray):
        g=nx.to_networkx_graph(g)    # if matrix is passed, convert to networkx
    d=dict(g.degree()).values()
    vals=list(set(d))
    counts=[d.count(i) for i in vals]
    return list(zip(vals, counts))

# return small world statistic of a graph
# returns metric of largest component if disconnected
def smallworld(a):
    g_sm=nx.from_numpy_matrix(a)
    g_sm=max(nx.connected_component_subgraphs(g_sm),key=len)   # largest component
    numnodes=g_sm.number_of_nodes()
    numedges=g_sm.number_of_edges()
    nodedegree=(numedges*2.0)/numnodes
    
    c_sm=nx.average_clustering(g_sm)        # c^ws in H&G (2006)
    #c_sm=sum(nx.triangles(usfg).values())/(# of paths of length 2) # c^tri
    l_sm=nx.average_shortest_path_length(g_sm)
    
    # c_rand same as edge density for a random graph? not sure if "-1" belongs in denominator, double check
    #c_rand= (numedges*2.0)/(numnodes*(numnodes-1))   # c^ws_rand?  
    c_rand= float(nodedegree)/numnodes                  # c^tri_rand?
    l_rand= np.log(numnodes)/np.log(nodedegree)    # approximation, see humphries & gurney (2008) eq 11
    #l_rand= (np.log(numnodes)-0.5772)/(np.log(nodedegree)) + .5 # alternative ASPL from fronczak, fronczak & holyst (2004)
    s=(c_sm/c_rand)/(l_sm/l_rand)
    return s




from . import *

# stochastic search through graph by node degree (proxy for frequency)
def nodeDegreeSearch(g, td):
    numnodes=nx.number_of_nodes(g)
    if td.trim <= 1:
        numtrim=int(round(numnodes*td.trim))       # if <=1, paramater is proportion of a list
    else:
        numtrim=td.trim                            # else, parameter is length of a list
    
    # make list of nodes by frequency
    l=[]
    for i, j in g.degree().items():
        l=l+[i]*j
    
    # simulate search
    walk=[]
    
    if td.start_node[0]=="specific":
        newnode=td.start_node[1]
        walk.append(newnode)
        l[:] = [j for j in l if j != newnode]
    
    while len(walk) < numtrim:
        newnode=np.random.choice(l)
        walk.append(newnode)
        l[:] = [j for j in l if j != newnode]
    
    return walk

# cluster-based depth first search: output all neighbors of starting node (random order), then all neighbors of most recently
# outputted node, etc; when you reach a dead end, back through the list until a new node with neighbors is usable
def cbdfs(g, td):
    import scipy    
    numnodes=nx.number_of_nodes(g)
    if td.trim <= 1:
        numtrim=int(round(numnodes*td.trim))       # if <=1, paramater is proportion of a list# make list of nodes by frequency
    else:
        numtrim=td.trim                            # else, parameter is length of a list
    # simulate search
    walk=[]
    
    if (td.start_node=="stationary") or (td.jumptype=="stationary"):
        a=np.array(nx.to_numpy_matrix(g))
        t=a/sum(a).astype(float)
        statdist=stationary(t)
        statdist=scipy.stats.rv_discrete(values=(list(range(len(t))),statdist))
    
    if td.start_node=="stationary":
        start=statdist.rvs(random_state=seed)      # choose starting point from stationary distribution #TODO: no definition of seed
    elif td.start_node=="uniform":
        start=np.random.choice(nx.nodes(g))        # choose starting point uniformly
    elif td.start_node[0]=="specific":
        start=td.start_node[1]
    
    unused_nodes=g.nodes()
    walk.append(start)
    unused_nodes.remove(start)
    currentnode=start
    
    # will have problems with disconnected graphs if numtrim is too high!
    while len(walk) < numtrim:
        next_nodes=g.neighbors(currentnode)
        next_nodes[:]=[i for i in next_nodes if i in unused_nodes]
        np.random.shuffle(next_nodes)
        if len(next_nodes) > 0:
            walk = walk + next_nodes
            unused_nodes[:] = [i for i in unused_nodes if i not in next_nodes]
            currentnode=walk[-1]
        else:
            currentnode = walk[walk.index(currentnode)-1]
    if len(walk) > numtrim:
        walk = walk[:numtrim]
    return walk

def spreadingActivationSearch(g, td, decay):
    import scipy    
    numnodes=nx.number_of_nodes(g)
    if td.trim <= 1:
        numtrim=int(round(numnodes*td.trim))       # if <=1, paramater is proportion of a list# make list of nodes by frequency
    else:
        numtrim=td.trim                            # else, parameter is length of a list
    
    if (td.start_node=="stationary") or (td.jumptype=="stationary"):
        a=np.array(nx.to_numpy_matrix(g))
        t=a/sum(a).astype(float)
        statdist=stationary(t)
        statdist=scipy.stats.rv_discrete(values=(list(range(len(t))),statdist))
    
    if td.start_node=="stationary":
        start=statdist.rvs(random_state=seed)      # choose starting point from stationary distribution #TODO: no definition of seed
    elif td.start_node=="uniform":
        start=np.random.choice(nx.nodes(g))        # choose starting point uniformly
    elif td.start_node[0]=="specific":
        start=td.start_node[1]
    
    activations=dict.fromkeys(list(range(numnodes)), 0)

    activations[start]=1.0
    walk=[start]
    
    while len(walk) < numtrim:
        newacts=dict.fromkeys(list(range(numnodes)), 0)
        walknodes=[]
        for node in range(numnodes):
            nn=g.neighbors(node)
            newact=activations[node]*decay
            for i in nn:
                newact += activations[i]
            if newact > 1.0:
                newact = 1.0
            newacts[node] = newact
        
        for i in activations:            # batch update
            activations[i]=newacts[i]
         
        denom = float(sum([activations[i] for i in activations if i not in walk]))
        probs=[activations[i]/denom for i in activations]
        for i in walk:
            probs[i]=0.0

        newnode=np.random.choice(list(range(numnodes)),p=probs)
        walk.append(newnode)
        activations[newnode]=1.0
        
    return walk



from . import *

# given list of cluster lengths, compute average cluster size of each list, then return avearge of that
# also works on single list
def clusterSize(l, scheme, clustertype='fluid'):
    clist = findClusters(l, scheme, clustertype)
    
    avglists=[]
    for i in clist:
        avglist=[]
        for l in i:
            avglist.append(np.mean(l))
        avglists.append(np.mean(avglist))
    return avglists

# given list of cluster lengths, compute average number of cluster switches of each list, then return avearge of that
# also works on single list
def clusterSwitch(l, scheme, clustertype='fluid'):
    clist = findClusters(l, scheme, clustertype)
    
    avglists=[]
    for i in clist:
        avgnum=[]
        if len(i) > 0:
            if isinstance(i[0], list):
                for l in clist:
                    avgnum.append(len(l)-1)
                avglists.append(np.mean(avgnum))
            else:
                avglists.append(len(i)-1)
        else:
            avglists.append(0)
    return avglists

# report average cluster size for list or nested lists
def findClusters(l, scheme, clustertype='fluid'):
    # only convert items to labels if list of items, not list of lists
    if len(l) > 0:
        if isinstance(l[0], list):
            clusters=l
        else:
            clusters=labelClusters(l, scheme)
    else:
        clusters=[]
    
    csize=[]
    curcats=set([])
    runlen=0
    clustList=[]
    firstitem=1
    for inum, item in enumerate(clusters):
        if isinstance(item, list):
            clustList.append(findClusters(item, scheme, clustertype=clustertype))
        else:
            newcats=set(item.split(';'))
            if newcats.isdisjoint(curcats) and firstitem != 1:      # end of cluster, append cluster length
                csize.append(runlen)
                runlen = 1
            else:                                                   # shared cluster or start of list
                runlen += 1
            
            if clustertype=="fluid":
                curcats = newcats
            elif clustertype=="static":
                curcats = (curcats & newcats)
                if curcats==set([]):
                    curcats = newcats
            else:
                raise ValueError('Invalid cluster type')
        firstitem=0
    csize.append(runlen)
    if sum(csize) > 0:
        clustList += csize
    return clustList

# returns labels in place of items for list or nested lists
# provide list (l) and coding scheme (external file)
def labelClusters(l, scheme, labelIntrusions=False, targetLetter=None):
    if isinstance(scheme,str):
        clustertype = "semantic"    # reads clusters from a fixed file
    elif isinstance(scheme,int):
        clustertype = "letter"      # if an int is given, use the first N letters as a clustering scheme
        maxletters = scheme
        if targetLetter:
            targetLetter = targetLetter.lower()
        
    else:
        raise Exception('Unknown clustering type in labelClusters()')

    if clustertype == "semantic":
        cf=open(scheme,'rt', encoding='utf-8-sig')
        cats={}
        for line in cf:
            line=line.rstrip()
            if line[0] == "#": continue         # skip commented lines
            cat, item = line.split(',')
            cat=cat.lower().replace(' ','').replace("'","").replace("-","") # basic clean-up
            item=item.lower().replace(' ','').replace("'","").replace("-","")
            if item not in list(cats.keys()):
                cats[item]=cat
            else:
                if cat not in cats[item]:
                    cats[item]=cats[item] + ';' + cat
    labels=[]
    for inum, item in enumerate(l):
        if isinstance(item, list):
            labels.append(labelClusters(item, scheme, labelIntrusions=labelIntrusions, targetLetter=targetLetter))
        else:
            item=item.lower().replace(' ','')
            if clustertype == "semantic":
                if item in list(cats.keys()):
                    labels.append(cats[item])
                elif labelIntrusions:               # if item not in dict, either ignore it or label is as category "intrusion"
                    labels.append("intrusion")
            elif clustertype == "letter":
                if item[0] == targetLetter:
                    labels.append(item[:maxletters])
                elif labelIntrusions:
                    labels.append("intrusion")
    return labels

def intrusionsList(l, scheme):  
    if len(l) > 0:
        if isinstance(l[0][0], list):
            intrusion_items = [intrusionsList(i, scheme) for i in l]
        else:
            if len(scheme) == 1:
                labels = labelClusters(l, 1, labelIntrusions=True, targetLetter=scheme)
            else:
                labels = labelClusters(l, scheme, labelIntrusions=True)
            intrusion_items = [[l[listnum][i] for i, j in enumerate(eachlist) if j=="intrusion"] for listnum, eachlist in enumerate(labels)]
    else:
        intrusion_items = []
    return intrusion_items

def intrusions(l, scheme):
    ilist = intrusionsList(l, scheme)
    
    # if fluency data are hierarchical, report mean per individual
    if isinstance(l[0][0], list):
        return [np.mean([len(i) for i in subj]) for subj in ilist]
    # if fluency data are non-hierarchical, report mean per list
    else:
        return [float(len(i)) for i in ilist]

def perseverationsList(l):
    if len(l) > 0:
        if isinstance(l[0][0], list):
            perseveration_items = [perseverationsList(i) for i in l]
        else:
            perseveration_items = [list(set([item for item in ls if ls.count(item) > 1])) for ls in l]
    else:
        perseveration_items = []
    return perseveration_items


def perseverations(l):
    def processList(l2):
        return [float(len(i)-len(set(i))) for i in l2]
    
    # if fluency data are hierarchical, report mean per individual
    if isinstance(l[0][0],list):
        return [np.mean(processList(subj)) for subj in l]
    # if fluency data are non-hierarchical, report mean per list
    else:
        return processList(l)



# these functions need to be made more extensible

from . import *

def commonNode(graph, items, node1, node2, numsims=100, jumpval=0.0):
    import scipy.sparse as sp
    
    if sp.issparse(graph):
        sparse=True
    else:
        sparse=False
    
    def chooseRandomNeighbor(pos):
        if sparse:
            return np.random.choice(sp.find(graph[pos])[1])
        else:
            return np.random.choice(np.where(graph[pos]==1)[0])

    inv_items = {v: k for k, v in items.items()}
    common_items={}
    for sim in range(numsims):
        a_idx = a_cur = inv_items[node1]
        b_idx = b_cur = inv_items[node2]
        idx=[a_idx, b_idx]
        a_hits=[]
        b_hits=[]
        intersect=[]
        while len(intersect) == 0:
            if np.random.random() < jumpval:
                a_cur = a_idx
            else:
                a_cur = chooseRandomNeighbor(a_cur)
            if np.random.random() < jumpval:
                b_cur = b_idx
            else:
                b_cur = chooseRandomNeighbor(b_cur)
            
            if a_cur not in a_hits and a_cur not in idx:
                a_hits.append(a_cur)
            if b_cur not in b_hits and b_cur not in idx:
                b_hits.append(b_cur)
            
            intersect = [val for val in a_hits if val in b_hits]

        incr_val = 1.0/len(intersect)
        for i in intersect:
            if items[i] not in common_items:
                common_items[items[i]] = incr_val
            else:
                common_items[items[i]] += incr_val
    
    for i in common_items:
        common_items[i] = common_items[i] / float(numsims)

    common_items = [(i,common_items[i]) for i in common_items]
    common_items = sorted(common_items, key=lambda tup: tup[1]) 

    return common_items

# first attempt to find similar items between node1 and node2 -- i think something is conceptually wrong here
#def commonNode(graph, items, node1, node2):
#    inv_items = {v: k for k, v in items.iteritems()}
#    t=graph/sum(graph.astype(float))                                                          # link to transition matrix
#    node_idx=[inv_items[node1], inv_items[node2]]                                             # item labels to ids
#    x=[i for i in range(len(graph)) if i not in node_idx]                                     # item ids to transition across freely
#    x1 = x + [node_idx[0], node_idx[1]]
#    x2 = x + [node_idx[1], node_idx[0]]
#    x1=np.array(x1)
#    x2=np.array(x2)
#    t1=t[x1[:,None],x1]                                                                       # re-arrange transition matrix to be in this order
#    t2=t[x2[:,None],x2]
#    
#    Q1=t1[:-1,:-1]
#    Q2=t2[:-1,:-1]
#    
#    # this can probably be made more efficient by using linalg instead of inv -- model after snafu.probX()
#    reg=(1+1e-10)                        
#    identmat=np.identity(len(graph)) * reg
#    I=identmat[:len(Q1),:len(Q1)]
#    N1=inv(I-Q1)
#    N2=inv(I-Q2)
#    #prob1=N1[-1,:-1]/sum(N1[-1,:-1])    # probability of starting at node 2 passing through node i before being absorbed by node 2
#    #prob2=N2[-1,:-1]/sum(N2[-1,:-1])    # vice versa
#    prob1=N1[-1,:-1]    # probability of starting at node 2 passing through node i before being absorbed by node 2
#    prob2=N2[-1,:-1]    # vice versa
#    #probs = np.multiply(prob1, prob2)
#    probs = np.add(prob1, prob2)
#    probs = probs/sum(probs)
#    similarityscore = sum(probs)
#    labels = [items[i] for i in x]
#    common = zip(probs, labels)
#    common = sorted(common, key=lambda tup: tup[0]) 
#
#    return zip(probs, labels), similarityscore


# monte carlo implementation of triadic comparison task, for when network is too large (works with sparse matrices)
# sloppy code put together quickly...
def triadicMonteCarlo(graph, items, triad, numsims=100,jumpval=0.0):
    import scipy.sparse as sp

    if sp.issparse(graph):
        sparse=True
    else:
        sparse=False
    
    def chooseRandomNeighbor(pos):
        if sparse:
            return np.random.choice(sp.find(graph[pos])[1])
        else:
            return np.random.choice(np.where(graph[pos]==1)[0])

    inv_items = {v: k for k, v in items.items()}
    ab = ac = bc = 0
    for sim in range(numsims):
        a_idx = a_cur = inv_items[triad[0]]
        b_idx = b_cur = inv_items[triad[1]]
        c_idx = c_cur = inv_items[triad[2]]
        fin=0
        while fin==0:
            if np.random.random() < jumpval:
                a_cur = a_idx
            else:
                a_cur = chooseRandomNeighbor(a_cur)
            if np.random.random() < jumpval:
                b_cur = b_idx
            else:
                b_cur = chooseRandomNeighbor(b_cur)
            if np.random.random() < jumpval:
                c_cur = c_idx
            else:
                c_cur = chooseRandomNeighbor(c_cur)
            
            # split wins
            numwins = 0
            if a_cur in [b_idx, c_idx]:
                numwins += 1
            if b_cur in [a_idx, c_idx]:
                numwins += 1
            if c_cur in [a_idx, b_idx]:
                numwins += 1
            if numwins > 1:
                # reset walk if there's a tie (could split wins, but then results are slightly different than analytic version)
                a_cur = a_idx
                b_cur = b_idx
                c_cur = c_idx

            
            if a_cur == b_idx:
                ab += 1.0
                fin = 1
            if a_cur == c_idx:
                ac += 1.0
                fin = 1
            if b_cur == a_idx:
                ab += 1.0
                fin = 1
            if b_cur == c_idx:
                bc += 1.0
                fin = 1
            if c_cur == a_idx:
                ac += 1.0
                fin = 1
            if c_cur == b_idx:
                bc += 1.0
                fin = 1
    numsims=float(numsims)
    return [ab/numsims, ac/numsims, bc/numsims]

def similarity(a, items, start, choices, steps="inf", jumpval=0.0):
    inv_items = {v: k for k, v in items.items()}
    t=a/sum(a.astype(float))                                                                # link to transition matrix
    choices_idx=[inv_items[i] for i in choices]                                             # item labels to ids
    x=[i for i in range(len(a)) if ((i not in choices_idx) and (i != inv_items[start]))]    # item ids to transition across freely
    x = x + [inv_items[start]] + choices_idx                                                # tack start state and absorbging states on the end
    x2=np.array(x)
    t2=t[x2[:,None],x2]                                                                       # re-arrange transition matrix to be in this order
   
    # indices/length for convenience later
    numchoices=len(choices)
    nonchoicelength = len(a)-numchoices
    start_idx = nonchoicelength-1
    choice_idx = list(range(start_idx+1,len(x)))
  
    # add jumps
    if jumpval > 0.0:
        t2 = t2 * (1.0-jumpval)
        t2[start_idx,:] = t2[start_idx,:] + jumpval        # + jumpval probability of transitioning to start node from any node
  
    # separate into Q (btw non-absorbing) and R (non-absorbing to absorbing)
    Q=t2[:nonchoicelength,:nonchoicelength]
    
    # compute absorbing probabilities
    if steps=="inf":
        # this can probably be made more efficient by using linalg instead of inv -- model after snafu.probX()
        reg=(1+1e-10)                        
        identmat=np.identity(len(a)) * reg
        I=identmat[:len(Q),:len(Q)]
        R=t2[nonchoicelength:,:nonchoicelength]
        N=inv(I-Q)
        B=np.dot(R,N)
        probs = B[:,nonchoicelength-1]
    elif steps >= 1:
        probs=[]
        newQ=np.zeros(nonchoicelength)
        newQ[start_idx]=1.0
        for stepnum in range(steps):
            ptmp=np.zeros(numchoices)
            for k in range(nonchoicelength):
                num1=newQ[k]
                for choicenum, choice in enumerate(choice_idx):
                    num2=t2[choice,k]                          # probability transitioning from k to absorbing node    
                    ptmp[choicenum] += (num1*num2)
            newQ=np.inner(newQ,Q)
            probs.append(ptmp)
    else:
        print("similarity() must take steps >=1 or 'inf'")
    
    return probs

def triadicComparison(graph, items, triad, steplimit=200, jumpval=0.0):
    # should simplify this code / extend to work with arbitrary number of items
    starta = similarity(graph, items, start=triad[0], choices=[triad[1],triad[2]],steps=steplimit, jumpval=jumpval)
    startb = similarity(graph, items, start=triad[1], choices=[triad[0],triad[2]],steps=steplimit, jumpval=jumpval)
    startc = similarity(graph, items, start=triad[2], choices=[triad[0],triad[1]],steps=steplimit, jumpval=jumpval)
    
    a_term = [sum(starta[0])]
    b_term = [sum(startb[0])]
    c_term = [sum(startc[0])]
    for step in range(1,steplimit):
        a_term.append(a_term[-1] + sum(starta[step]))   # probability that chain a has terminated by step X
        b_term.append(b_term[-1] + sum(startb[step]))   # " chain b
        c_term.append(c_term[-1] + sum(startc[step]))   # " chain c
    
    # e.g., startab = probability that chain starting at a ends at b before chains starting from b or c terminate
    startab = startac = startba = startbc = startca = startcb = 0
    for step in range(steplimit):
        startab += starta[step][0] * (1-b_term[step]) * (1-c_term[step])
        startac += starta[step][1] * (1-b_term[step]) * (1-c_term[step])
        startba += startb[step][0] * (1-a_term[step]) * (1-c_term[step])
        startbc += startb[step][1] * (1-a_term[step]) * (1-c_term[step])
        startca += startc[step][0] * (1-a_term[step]) * (1-b_term[step])
        startcb += startc[step][1] * (1-a_term[step]) * (1-b_term[step])
    
    denom = (startab+startac+startba+startbc+startca+startcb)
    choiceab = (startab + startba) / denom
    choiceac = (startac + startca) / denom
    choicebc = (startbc + startcb) / denom

    return [choiceab, choiceac, choicebc]




from . import *

# http://stackoverflow.com/a/32107024/353278
# use dot notation on dicts for convenience
class dotdict(dict):
    """
    Example:
    m = dotdict({'first_name': 'Eduardo'}, last_name='Pool', age=24, sports=['Soccer'])
    """
    def __init__(self, *args, **kwargs):
        super(dotdict, self).__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k, v in list(arg.items()):
                    self[k] = v

        if kwargs:
            for k, v in list(kwargs.items()):
                self[k] = v

    def __getattr__(self, attr):
        return self.get(attr)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        super(dotdict, self).__setitem__(key, value)
        self.__dict__.update({key: value})

    def __delattr__(self, item):
        self.__delitem__(item)

    def __delitem__(self, key):
        super(Map, self).__delitem__(key)  # TODO: no definition of Map
        del self.__dict__[key]

# from http://locallyoptimal.com/blog/2013/01/20/elegant-n-gram-generation-in-python/
# generate list of ngrams
def find_ngrams(input_list, n):
    return list(zip(*[input_list[i:] for i in range(n)]))

# modified from ExGUtils package by Daniel Gamermann <gamermann@gmail.com>
# helper function generate flast lists from nested lists
# modified from http://stackoverflow.com/a/952952/353278
# flattens list of list one level only, preserving non-list items
# flattens type list and type np.ndarray, nothing else (on purpose)
def flatten_list(l, numtimes=1):
    l1 = [item for sublist in l if isinstance(sublist,list) or isinstance(sublist,np.ndarray) for item in sublist]
    l = l1+[item for item in l if not isinstance(item,list) and not isinstance(item,np.ndarray)]
    if numtimes > 1:
        l = flatten_list(l, numtimes-1)
    return l

# log trick given list of log-likelihoods **UNUSED
def logTrick(loglist):
    logmax=max(loglist)
    loglist=[i-logmax for i in loglist]                     # log trick: subtract off the max
    p=np.log(sum([np.e**i for i in loglist])) + logmax  # add it back on
    return p

# helper function grabs highest n items from list items **UNUSED
# http://stackoverflow.com/questions/350519/getting-the-lesser-n-elements-of-a-list-in-python
def maxn(items,n):
    maxs = items[:n]
    maxs.sort(reverse=True)
    for i in items[n:]:
        if i > maxs[-1]: 
            maxs.append(i)
            maxs.sort(reverse=True)
            maxs= maxs[:n]
    return maxs

# find best ex-gaussian parameters
# port from R's retimes library, mexgauss function by Davide Massidda <davide.massidda@humandata.it>
# returns [mu, sigma, lambda]
def mexgauss(rts):
    n = len(rts)
    k = [np.nan, np.nan, np.nan]
    start = [np.nan, np.nan, np.nan]
    k[0] = np.mean(rts)
    xdev = [rt - k[0] for rt in rts]
    k[1] = sum([i**2 for i in xdev])/(n - 1.0)
    k[2] = sum([i**3 for i in xdev])/(n - 1.0)
    if (k[2] > 0):
        start[2] = (k[2]/2.0)**(1/3.0)
    else:
        start[2] = 0.8 * np.std(rts)
    start[1] = np.sqrt(abs(k[1] - start[2]**2))
    start[0] = k[0] - start[2]
    start[2] = (1.0/start[2])   # tau to lambda
    return(start)

# decorator; disables garbage collection before a function, enable gc after function completes
# provides some speed-up for functions that have lots of unnecessary garbage collection (e.g., lots of list appends)
def nogc(fun):
    import gc
    def gcwrapper(*args, **kwargs):
        gc.disable()
        returnval = fun(*args, **kwargs)
        gc.enable()
        return returnval
    return gcwrapper

# take list of lists in number/node and translate back to items using dictionary (e.g., 1->dog, 2->cat)
def numToItemLabel(data, items):
    new_data=[]
    for l in data:
        new_data.append([])
        for i in l:
            new_data[-1].append(items[i])
    return new_data

# modified from ExGUtils package by Daniel Gamermann <gamermann@gmail.com>
def rand_exg(irt, sigma, lambd):
    tau=(1.0/lambd)
    nexp = -tau*np.log(1.-np.random.random())
    ngau = np.random.normal(irt, sigma)
    return nexp + ngau

#def renumber(Xs,numsubs,numper):
#    start=0
#    end=numper
#    ssnumnodes=[]
#    itemsb=[]
#    datab=[]
#    for sub in range(len(subs)):
#        subXs = Xs[start:end]
#        itemset = set(snafu.flatten_list(subXs))
#        ssnumnodes.append(len(itemset))
#                                                    
#        ss_items = {}
#        convertX = {}
#        for itemnum, item in enumerate(itemset):
#            ss_items[itemnum] = items[item]
#            convertX[item] = itemnum
#                                                    
#        itemsb.append(ss_items)
#                                                    
#        subXs = [[convertX[i] for i in x] for x in subXs]
#        datab.append(subXs)
#        start += 3
#        end += 3

# decorator; prints elapsed time for function call
def timer(fun):
    from datetime import datetime
    def timerwrapper(*args, **kwargs):
        starttime=datetime.now()
        returnval = fun(*args, **kwargs)
        elapsedtime=str(datetime.now()-starttime)
        print(elapsedtime)
        return returnval
    return timerwrapper

def reverseDict(items):
    newitems=dict()
    for itemnum in items:
        itemlabel = items[itemnum]
        newitems[itemlabel] = itemnum
    return newitems

# remove perseverations -- keep only first occurrence in place
# https://www.peterbe.com/plog/uniqifiers-benchmark
def no_persev(x):
    seen = set()
    seen_add = seen.add
    return [i for i in x if not (i in seen or seen_add(i))]

# this function is copied from scipy to avoid shipping that whole library with snafu
# unlike scipy version, this one doesn't return p-value (requires C code from scipy)
def pearsonr(x, y):
    
    def _sum_of_squares(a, axis=0):
        a, axis = _chk_asarray(a, axis)
        return np.sum(a*a, axis)

    def _chk_asarray(a, axis):
        if axis is None:
            a = np.ravel(a)
            outaxis = 0
        else:
            a = np.asarray(a)
            outaxis = axis

        if a.ndim == 0:
            a = np.atleast_1d(a)

        return a, outaxis
    
    
    # x and y should have same length.
    x = np.asarray(x)
    y = np.asarray(y)
    n = len(x)
    mx = x.mean()
    my = y.mean()
    xm, ym = x - mx, y - my
    r_num = np.add.reduce(xm * ym)
    r_den = np.sqrt(_sum_of_squares(xm) * _sum_of_squares(ym))
    r = r_num / r_den

    return r

# takes an individual's data in group space and translates it into local space
def groupToIndividual(Xs, group_dict):
    itemset = set(flatten_list(Xs))
    
    ss_items = {}
    convertX = {}
    for itemnum, item in enumerate(itemset):
        ss_items[itemnum] = group_dict[item]
        convertX[item] = itemnum
    
    Xs = [[convertX[i] for i in x] for x in Xs]
    
    return Xs, ss_items

# take Xs and convert them from numbers (nodes) to labels
def numToLabel(Xs, items):
    for lnum, l in enumerate(Xs):
        for inum, i in enumerate(l):
            Xs[lnum][inum]=items[i]
    return Xs

# flat list from tuple walk
def nodes_from_edges(walk):
    path=list(list(zip(*walk))[0]) # first element from each tuple
    path.append(walk[-1][1]) # second element from last tuple
    return path

# tuple walk from flat list
def edges_from_nodes(path):
    walk=[]
    for i in range(len(path)-1):
        walk.append((path[i],path[i+1])) 
    return walk

def stationary(t, method="unweighted"):
    if method=="unweighted":                 # only works for unweighted matrices!
        return sum(t>0)/float(sum(sum(t>0)))
    else:                                       # buggy
        eigen=np.linalg.eig(t)[1][:,0]
        return np.real(eigen/sum(eigen))


# Unique nodes in random walk preserving order
# (aka fake participant data)
# http://www.peterbe.com/plog/uniqifiers-benchmark
def censored(walk, td=None, seed=None):
    def addItem(item):
        seen[item] = 1
        result.append(item)

    nplocal = np.random.RandomState(seed)
    seen = {}
    result = []
    for item in nodes_from_edges(walk):
        if item in seen:
            try:
                if nplocal.rand() <= td.censor_fault:
                    addItem(item)
            except:
                continue
        else:
            try:
                if nplocal.rand() <= td.emission_fault:
                    continue
                else:
                    addItem(item)
            except:
                addItem(item)
    return result

# first hitting times for each node
# TODO: Doesn't work with faulty censoring!!!
def firstHits(walk):
    firsthit=[]
    path=edges_from_nodes(walk)
    for i in censored(walk):
        firsthit.append(path.index(i))
    return list(zip(censored(walk),firsthit))



import pickle
import networkx as nx
import numpy as np
#import scipy.cluster
#import scipy.stats
#import scipy
import operator
import math
import sys
import copy
import csv
import warnings
import json
import os
import sys


from numpy.linalg import inv
from itertools import *
from datetime import datetime

from ._version import __version__

from .helper import *
from .structs import *
from .io import *
from .generate_lists import *
from .netstats import *
from .clustering import *
from .frequency import *
from .search import *
from .triadic import *
from .word_properties import *
from .pci import *
from .generate_graphs import *
from .irts import *

from .core import *
from . import gui



from . import *

# returns a vector of how many hidden nodes to expect between each Xi for each X in Xs
def expectedHidden(Xs, a):
    numnodes=len(a)
    expecteds=[]
    t=a/sum(a.astype(float))                      # transition matrix (from: column, to: row)
    identmat=np.identity(numnodes) * (1+1e-10)    # pre-compute for tiny speed-up
    for x in Xs:
        x2=np.array(x)
        t2=t[x2[:,None],x2]                       # re-arrange transition matrix to be in list order
        expected=[]
        for curpos in range(1,len(x)):
            Q=t2[:curpos,:curpos]
            I=identmat[:len(Q),:len(Q)]
            N=np.linalg.solve(I-Q,I[-1])
            expected.append(sum(N))
            #N=inv(I-Q)         # old way, a little slower
            #expected.append(sum(N[:,curpos-1]))
        expecteds.append(expected)        
    return expecteds

# generates fake IRTs from # of steps in a random walk, using gamma distribution
def stepsToIRT(irts, seed=None):
    nplocal=np.random.RandomState(seed)        # to generate the same IRTs each time
    new_irts=[]
    for irtlist in irts.data:
        if irts.irttype=="gamma":
            newlist=[nplocal.gamma(irt, (1.0/irts.gamma_beta)) for irt in irtlist]  # beta is rate, but random.gamma uses scale (1/rate)
        if irts.irttype=="exgauss":
            newlist=[rand_exg(irt, irts.exgauss_sigma, irts.exgauss_lambda) for irt in irtlist] 
        new_irts.append(newlist)
    return new_irts

# ** this function is not really needed anymore since moving functionality to genX, 
# ** but there may be some niche cases where needed...
# trim Xs to proportion of graph size, the trim graph to remove any nodes that weren't hit
# used to simulate human data that doesn't cover the whole graph every time
def trim_lists(trimprop, Xs, steps):
    numnodes=len(Xs[0])             # since Xs haven't been trimmed, we know list covers full graph
    alter_graph_size=0              # report if graph size changes-- may result in disconnected graph!

    if trimprop <= 1:
        numtrim=int(round(numnodes*trimprop))       # if <=1, paramater is proportion of a list
    else:
        numtrim=trimprop                            # else, parameter is length of a list

    Xs=[i[0:numtrim] for i in Xs]
    steps=[i[0:(numtrim-1)] for i in steps]
    for i in range(numnodes):
        if i not in set(flatten_list(Xs)):
            alter_graph_size=1
    return Xs, steps, alter_graph_size




from . import *

def list_subjects_and_categories(command, root_path):
    subjects=[]
    categories=[]
    groups=["all"]
    
    with open(command['fullpath'],'rt',encoding="utf-8-sig") as fh:
        header=fh.readline().strip().split(',')
        subj_idx = header.index("id")
        cat_idx = header.index("category")
        try:
            group_idx = header.index("group")
        except:
            group_idx = -1

        from csv import reader
        for line in reader(fh):
            if line[subj_idx] not in subjects:
                subjects.append(line[subj_idx])
            if line[cat_idx] not in categories:
                categories.append(line[cat_idx])
            if group_idx > -1:
                group_label = line[group_idx]
                if group_label not in groups:
                    groups.append(group_label)
    return { "type": "list_subjects_and_categories",
             "subjects": subjects, 
             "categories": categories,
             "groups": groups,
             "subject": subjects[0],
             "category": categories[0],
             "group": groups[0] }

def jsonGraph(g, items):
    json_data = nx.readwrite.json_graph.node_link_data(g)
    
    json_data['edges'] = json_data['links']
    json_data.pop('links', None)
    json_data.pop('directed', None)
    json_data.pop('multigraph', None)
    json_data.pop('graph', None)
    
    for i, j in enumerate(json_data['edges']):
        json_data['edges'][i]['id'] = i
        
        # this line fixes a bug (JSON not serializable) i don't know why it's not serializable without it (py3.5)
        json_data['edges'][i]['target'] = int(json_data['edges'][i]['target']) 
   
    for i, j in enumerate(json_data['nodes']):
        json_data['nodes'][i]['label'] = items[j['id']]
   
    # works when edges['target'] aren't in dict
    #json_data.pop('edges',None)

    return json_data

def label_to_filepath(x, root_path, filetype):
    filedict=dict()
    folder = root_path + "/" + filetype + "/"

    # since filenames are populated from directory listing there's no need to build a dict() anymore, but that's the way it's done still
    for filename in os.listdir(folder):
        if "csv" in filename:
            label = filename[0:filename.find('.')].replace('_',' ')
            filedict[label] = folder + filename
    try:
        filename = filedict[x]
    except:
        filename = None
    return filename

def data_properties(command, root_path):
   
    # turns array into string: "Mean (Std) [Min - Max]"
    def format_output(x):
        x_mean = str(round(np.mean(x),2))
        x_std = str(round(np.std(x),2))
        x_min = str(round(np.min(x),2))
        x_max = str(round(np.max(x),2))
        
        return x_mean + " (" + x_std + ") [" + x_min + " -- " + x_max + "]"
    
    command = command['data_parameters']
  
    if command['factor_type'] == "subject":
        subject = str(command['subject'])
        group = None
    elif command['factor_type'] == "group":
        if command['group'] != "all":
            group = command['group']
        else:
            group = None                             # reserved group label in GUI for all subjects
        subject = None

    filedata = load_fluency_data(command['fullpath'], category=command['category'], spell=label_to_filepath(command['spellfile'], root_path, "spellfiles"), group=group, subject=subject, hierarchical=True)
    Xs = filedata.Xs
    labeledXs = filedata.labeledXs
    items = filedata.items
    irts = filedata.irts
    numnodes = filedata.numnodes

    # initialize
    avg_cluster_size = ["n/a"]
    avg_num_cluster_switches = ["n/a"]
    avg_num_intrusions = ["n/a"]
    num_lists = []
    avg_items_listed = []
    avg_unique_items_listed = []
    list_of_intrusions = []
    avg_num_perseverations = []
    list_of_perseverations = []
   
    if not command['freq_ignore']:
        try:
            freq_sub = float(command['freq_sub'])
        except:
            freq_sub = None
    else:
        freq_sub = None

    if not command['aoa_ignore']:
        try:
            aoa_sub = float(command['aoa_sub'])
        except:
            aoa_sub = None
    else:
        aoa_sub = None


    freqfile = label_to_filepath(command['freqfile'], root_path,"frequency")
    aoafile = label_to_filepath(command['aoafile'], root_path,"aoa")
    
    preset_schemes = {"Phonemic: 1 letter": 1,
                      "Phonemic: 2 letters": 2,
                      "Phonemic: 3 letters": 3}
    if command['cluster_scheme'] in preset_schemes.keys():
        schemefile = preset_schemes[command['cluster_scheme']]
    else:
        schemefile = label_to_filepath(command['cluster_scheme'], root_path, "schemes")
    
    total_words = 0
    avg_word_freq = []
    word_freq_excluded = []
    avg_word_aoa = []
    word_aoa_excluded = []

    if command['cluster_scheme'] != "None":
        avg_cluster_size = clusterSize(labeledXs, schemefile, clustertype=command['cluster_type'])
        avg_num_cluster_switches = clusterSwitch(labeledXs, schemefile, clustertype=command['cluster_type'])
        avg_num_intrusions = intrusions(labeledXs, schemefile)
        list_of_intrusions = intrusionsList(labeledXs, schemefile)
    avg_num_perseverations = perseverations(labeledXs)
    list_of_perseverations = perseverationsList(labeledXs)

    for subjnum in range(len(labeledXs)):
        num_lists.append(len(labeledXs[subjnum]))
        avg_items_listed.append(np.mean([len(i) for i in labeledXs[subjnum]]))
        avg_unique_items_listed.append(np.mean([len(set(i)) for i in labeledXs[subjnum]]))

        tmp1, tmp2 = wordFrequency(labeledXs[subjnum],missing=freq_sub,data=freqfile)
        avg_word_freq.append(np.mean(tmp1))
        for i in tmp2:
            word_freq_excluded.append(i)
        tmp1, tmp2 = ageOfAquisition(labeledXs[subjnum],missing=aoa_sub,data=aoafile)
        avg_word_aoa.append(np.mean(tmp1))
        for i in tmp2:
            word_aoa_excluded.append(i)
        for i in labeledXs[subjnum]:
            total_words += len(i)

    # clean up / format data to send back, still messy
    list_of_intrusions = flatten_list(list_of_intrusions)
    list_of_perseverations = flatten_list(list_of_perseverations)

    if len(labeledXs) > 1:
        if command['cluster_scheme'] != "None":
            avg_cluster_size = format_output(avg_cluster_size)
            avg_num_cluster_switches = format_output(avg_num_cluster_switches)
            avg_num_intrusions = format_output(avg_num_intrusions)
            avg_num_perseverations = format_output(avg_num_perseverations)
        num_lists = format_output(num_lists)
        avg_items_listed = format_output(avg_items_listed)
        avg_unique_items_listed = format_output(avg_unique_items_listed)
        avg_word_freq=format_output(avg_word_freq)
        avg_word_aoa=format_output(avg_word_aoa)
 
    word_freq_rate = 0
    for i in word_freq_excluded:
        word_freq_rate += len(i)
    word_freq_rate = str(round(float(word_freq_rate)/total_words*100,2))+'%'
    word_aoa_rate = 0
    for i in word_aoa_excluded:
        word_aoa_rate += len(i)
    word_aoa_rate = str(round(float(word_aoa_rate)/total_words*100,2))+'%'
    # fix
    csv_file = generate_csv_file(command, root_path)

    return { "type": "data_properties", 
             "num_lists": num_lists,
             "avg_items_listed": avg_items_listed,
             "intrusions": list_of_intrusions,
             "perseverations": list_of_perseverations,
             "avg_num_intrusions": avg_num_intrusions,
             "avg_num_perseverations": avg_num_perseverations,
             "avg_unique_items_listed": avg_unique_items_listed,
             "avg_num_cluster_switches": avg_num_cluster_switches,
             "avg_cluster_size": avg_cluster_size,
             "avg_word_freq": avg_word_freq,
             "avg_word_aoa": avg_word_aoa,
             "word_freq_rate": word_freq_rate,
             "word_freq_excluded": word_freq_excluded,
             "word_aoa_rate": word_aoa_rate,
             "word_aoa_excluded": word_aoa_excluded,
             "csv_file": csv_file }

# broken
def generate_csv_file(command, root_path):
    return ### temporary, function is broken JZ
    csv_file = "id,listnum,num_items_listed,num_unique_items,num_cluster_switches,avg_cluster_size,num_intrusions,num_perseverations,avg_word_freq,avg_word_aoa\n"
    filedata = load_fluency_data(command['fullpath'],category=command['category'], scheme=label_to_filepath(command['cluster_scheme'], root_path, "schemes"), spellfile=label_to_filepath(command['spellfile'], root_path, "spellfiles"))
    filedata.hierarchical()
    
    for subnum, sub in enumerate(filedata.subs):
        labeledXs = filedata.labeledXs[subnum]
        for listnum in range(len(filedata.Xs[subnum])):
            csv_sub = sub
            csv_listnum = listnum
            csv_numitems = len(filedata.Xs[subnum][listnum])
            csv_uniqueitem = len(set(filedata.Xs[subnum][listnum]))
            
            # parameters should come from snafu gui (scheme, clustertype)
            csv_clusterlength = clusterSize(labeledXs, scheme=label_to_filepath(command['cluster_scheme'], root_path, "schemes"), clustertype=command['cluster_type'])
            csv_clusterswitch = clusterSwitch(labeledXs, scheme=label_to_filepath(command['cluster_scheme'], root_path, "schemes"), clustertype=command['cluster_type'])

            # parameters should come from snafu gui (scheme)
            csv_intrusions = intrusions(labeledXs,scheme=label_to_filepath(command['cluster_scheme'], root_path, "schemes"))
            csv_perseverations = perseverations(labeledXs)

            csv_freq, temp = wordFrequency([labeledXs[listnum]],freq_sub=float(command['freq_sub']))
            csv_aoa, temp = ageOfAquisition([labeledXs[listnum]])
            csv_freq = np.mean(csv_freq)
            csv_aoa = np.mean(csv_aoa)

            csv_file += str(csv_sub)+','+str(csv_listnum)+','+str(csv_numitems)+','+str(csv_uniqueitem)+','+str(csv_clusterswitch)+','+str(round(csv_clusterlength,2))+','+str(csv_intrusions)+','+str(csv_perseverations)+','+str(round(csv_freq,2))+','+str(round(csv_aoa,2))+'\n'

    return csv_file


def network_properties(command, root_path):
    subj_props = command['data_parameters']
    command = command['network_parameters']

    # U-INVITE won't work with perseverations
    if command['network_method'] == "U-INVITE":
        removePerseverations=True
    else:
        removePerseverations=False
   
    if command['factor_type'] == "subject":
        subject = str(command['subject'])
        group = None
    elif command['factor_type'] == "group":
        if command['group'] != "all":
            group = command['group']
        else:
            group = None                             # reserved group label in GUI for all subjects
        subject = None

    filedata = load_fluency_data(subj_props['fullpath'], category=subj_props['category'], spell=label_to_filepath(subj_props['spellfile'], root_path, "spellfiles"), removePerseverations=removePerseverations, subject=subject, group=group)
    Xs = filedata.Xs
    items = filedata.items
    irts = filedata.irts
    numnodes = filedata.numnodes
    
    toydata=DataModel({
            'numx': len(Xs),
            'trim': 1,
            'jump': float(command['jump_probability']),
            'jumptype': command['jump_type'],
            'priming': float(command['priming_probability']),
            'start_node': command['first_item']})
    fitinfo=Fitinfo({
            'prior_method': "zeroinflatedbetabinomial",
            'prior_a': 1,
            'prior_b': 2,
            'zibb_p': 0.5,
            'startGraph': command['starting_graph'],
            'cn_windowsize': int(command['cn_windowsize']),
            'cn_threshold': int(command['cn_threshold']),
            'cn_alpha': float(command['cn_alpha']),
            'followtype': "avg", 
            'prune_limit': 100,
            'triangle_limit': 100,
            'other_limit': 100})
   
    if command['prior']=="None":
        prior=None
    elif command['prior']=="USF":
        usf_file_path = "/snet/USF_animal_subset.snet"
        filename = root_path + usf_file_path
        
        usf_graph, usf_items = read_graph(filename)
        usf_numnodes = len(usf_items)
        priordict = genGraphPrior([usf_graph], [usf_items], fitinfo=fitinfo)
        prior = (priordict, usf_items)
        
    if command['network_method']=="Naive Random Walk":
        bestgraph = naiveRandomWalk(Xs, numnodes=numnodes)
    elif command['network_method']=="Community Network":
        bestgraph = communityNetwork(Xs, fitinfo=fitinfo, numnodes=numnodes)
    elif command['network_method']=="Pathfinder":
        bestgraph = pathfinder(Xs, numnodes=numnodes)
    elif command['network_method']=="Correlation-based Network":
        bestgraph = correlationBasedNetwork(Xs, numnodes=numnodes)
    elif command['network_method']=="First Edge":
        bestgraph = firstEdge(Xs, numnodes=numnodes)
    elif command['network_method']=="U-INVITE":
        bestgraph, ll = uinvite(Xs, toydata, numnodes=numnodes, fitinfo=fitinfo, debug=False, prior=prior)
    
    nxg = nx.to_networkx_graph(bestgraph)
    nxg_json = jsonGraph(nxg, items)
    
    return graph_properties(nxg,nxg_json)

def analyze_graph(command, root_path): # used when importing graphs
    nxg_json = json.load(open(command['fullpath'],'rt',encoding="utf-8-sig"))
    nxg = nx.readwrite.nx.readwrite.json_graph.node_link_graph(
        nxg_json,
        multigraph = False,
        attrs=dict(source='source', target='target', name='id', key='nodes', link='edges')
    )
    return graph_properties(nxg, nxg_json)

def graph_properties(nxg,nxg_json): # separate function that calculates graph properties
    node_degree = np.mean(list(dict(nxg.degree()).values()))
    clustering_coefficient = nx.average_clustering(nxg)
    try:
        aspl = nx.average_shortest_path_length(nxg)
    except:
        aspl = "disjointed graph"
    density = nx.classes.function.density(nxg)
    betweenness_centrality_nodes = nx.algorithms.centrality.betweenness_centrality(nxg)
    avg_betweenness_centrality = 0.0
    for node, bc in list(betweenness_centrality_nodes.items()):
        avg_betweenness_centrality += bc
    avg_betweenness_centrality /= len(betweenness_centrality_nodes)

    return {
        "type": "network_properties",
        "node_degree": node_degree,
        "clustering_coefficient": clustering_coefficient,
        "aspl": aspl,
        "density": density,
        "avg_betweenness_centrality": avg_betweenness_centrality,
        "graph": nxg_json
    }

def quit(command, root_path): 
    return { "type": "quit",
             "status": "success" }

def error(msg):
    return { "type": "error",
             "msg": msg }



from setuptools import setup, find_packages
import re

# https://stackoverflow.com/a/7071358/353278
VERSIONFILE="snafu/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

setup(name='pysnafu',
      version=verstr,
      description='Generate semantic networks from fluency data',
      url='https://github.com/AusterweilLab/snafu-py',
      author='The Austerweil Lab at UW-Madison',
      author_email='jeffzemla@gmail.com',
      keywords=['fluency', 'networks','psychology','recall'],
      packages=['snafu'],
      include_package_data=True,
      install_requires=['numpy','networkx>=2.1','more_itertools'],
      zip_safe=False,
      classifiers=[
            'Programming Language :: Python :: 3.5'
      ]
      )



from . import *

def wordFrequency(subj,missing=0.5,data=None):
    return wordStat(subj,missing=missing,data=data)

def ageOfAquisition(subj,missing=None,data=None):
    return wordStat(subj,missing=missing,data=data)

def wordStat(subj,missing=None,data=None):
    # load dictionary
    d_val = {}
    with open(data, 'rt', encoding='utf-8-sig') as csvfile:
        # allows comments in file thanks to https://stackoverflow.com/a/14158869/353278
        reader = csv.DictReader(filter(lambda row: row[0]!='#', csvfile), fieldnames=['word','val'])
        for row in reader:
            d_val[row['word']]= float(row['val'])

    word_val = []
    words_excluded = []
    for i in subj: # each list
            temp=[]
            excluded=[]
            for j in i: # each word
                if (j in d_val): # word must be in the list
                    temp.append(d_val[j])
                else: # or their would be excluded
                    if (missing!=None): # case 2: not in the list, substituted by missing
                        temp.append(missing)
                    else:
                        excluded.append(j)
            if(len(temp)>0):
                word_val.append(np.mean(temp))
            words_excluded.append(excluded)
    return word_val, words_excluded




from . import *


# return simulated data on graph g
# also return number of steps between first hits (to use for IRTs)
def gen_lists(g, td, seed=None):
    Xs=[]
    steps=[]
    priming_vector=[]

    for xnum in range(td.numx):
        if seed == None:
            seedy = None
        else:
            seedy = seed + xnum
        rwalk=random_walk(g, td, priming_vector=priming_vector, seed=seedy)
        x=censored(rwalk, td)
        # fh=list(zip(*firstHits(rwalk)))[1]
        # step=[fh[i]-fh[i-1] for i in range(1,len(fh))]
        Xs.append(x)
        if td.priming > 0.0:
            priming_vector=x[:]
        # steps.append(step)
    td.priming_vector = []      # reset mutable priming vector between participants; JCZ added 9/29, untested

    alter_graph_size=0
    if td.trim != 1.0:
        numnodes=nx.number_of_nodes(g)
        for i in range(numnodes):
            if i not in set(flatten_list(Xs)):
                alter_graph_size=1

    return Xs, steps, alter_graph_size

# given an adjacency matrix, take a random walk that hits every node; returns a list of tuples
def random_walk(g, td, priming_vector=[], seed=None):
    import scipy.stats
    nplocal=np.random.RandomState(seed)    

    def jump():
        if td.jumptype=="stationary":
            second=statdist.rvs(random_state=seed)     # jump based on statdist
        elif td.jumptype=="uniform":
            second=nplocal.choice(nx.nodes(g))         # jump uniformly
        return second

    if (td.start_node=="stationary") or (td.jumptype=="stationary"):
        a=np.array(nx.to_numpy_matrix(g))
        t=a/sum(a).astype(float)
        statdist=stationary(t)
        statdist=scipy.stats.rv_discrete(values=(list(range(len(t))),statdist))
    
    if td.start_node=="stationary":
        start=statdist.rvs(random_state=seed)    # choose starting point from stationary distribution
    elif td.start_node=="uniform":
        start=nplocal.choice(nx.nodes(g))        # choose starting point uniformly
    elif td.start_node[0]=="specific":
        start=td.start_node[1]

    walk=[]
    unused_nodes=set(nx.nodes(g))
    unused_nodes.remove(start)
    first=start
    
    numnodes=nx.number_of_nodes(g)
    if td.trim <= 1:
        numtrim=int(round(numnodes*td.trim))       # if <=1, paramater is proportion of a list
    else:
        numtrim=td.trim                            # else, parameter is length of a list
    num_unused = numnodes - numtrim

    censoredcount=0                                # keep track of censored nodes and jump after td.jumponcensored censored nodes

    numsteps = 0
    while (len(unused_nodes) > num_unused) and ((td.maxsteps == None) or (numsteps < td.maxsteps)):       # covers td.trim nodes-- list could be longer if it has perseverations

        # jump after n censored nodes or with random probability (depending on parameters)
        if (censoredcount == td.jumponcensored) or (nplocal.random_sample() < td.jump):
            second=jump()
        else:                                           # no jumping!
            second=nplocal.choice([x for x in nx.all_neighbors(g,first)]) # follow random edge (actual random walk!)
            if (td.priming > 0.0) and (len(priming_vector) > 0):
                if (first in priming_vector[:-1]) & (nplocal.random_sample() < td.priming):      
                    idx=priming_vector.index(first)
                    second=priming_vector[idx+1]          # overwrite RW... kinda janky
        walk.append((first,second))
        numsteps += 1
        if second in unused_nodes:
            unused_nodes.remove(second)
            censoredcount=0
        else:
            censoredcount += 1
        first=second
    return walk
